import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from glob import glob
from tqdm import tqdm
from natsort import natsorted
from scipy.ndimage import distance_transform_edt
from .utils import load
import os
import matplotlib

def eval_angular_growth(folder, outfile=None):
    """
    Evaluates the growth in 5° angular sections (parelell to the agular pressure analyis) and the total growth of the 
    spheroid using the continuously tracked segmentation (therfore the "continous_segmentation" must be set to True
    in the function "compute_displacement_series")
    
    
    - folder: contains PIV results with segmentation
    - outfile: can be used to change the standard name of the excel file - if none excel files are created 
    in the given folder
    
    Returns:
    saves an excel file with the angular growth (angular area section at each timepoint / dived by 
                                                  angular area at timepoint 0)
    and another excel file with the total growth(total segmented area  at each timepoint / dived by 
                                                  total segmented area at timepoint 0)
    
    """
       
    ## load segmentations
    seg_files = natsorted(glob(folder+'/seg*.npy'))
       
    # initial spheroid mask 
    mask_t0 = load(seg_files[0])['mask']
    # intitial spheoir center for all further analysis
    cx_t0, cy_t0 = load(seg_files[0])['centroid']
    
    # Calculate Angle + Distances 
    y,x = np.indices(mask_t0.shape)
    dx = x - cx_t0
    dy = y - cy_t0
    #distance = np.sqrt(dx ** 2 + dy ** 2)  # dist to center
    angle = np.arctan2(dy, dx) 
           
    # initialize result dictionary
    results = {'total_growth': []}
    
    # create dict with all angles
    angles_dict = {str(a): [] for a in range(-180, 180, 5)}
    
      
    # loop over series of PIV segmentations
    for seg_file in tqdm(seg_files):
        # load current segmentation
        mask = load(seg_file)['mask']      
        ## total growth
        results['total_growth'].append(np.sum(mask)/np.sum(mask_t0))            
        # growth in angular section
        angular_sections = []
        growth_angular_sections = []
        for alpha in range(-180, 180, 5):
            mask2 = (angle >= (alpha-5)*np.pi/180.) & (angle < (alpha+5)*np.pi/180.)
            
           
            angular_sections.append(alpha)
            growth_angular_sections.append(np.nansum(mask[mask2])/np.nansum(mask_t0[mask2]))  #### growth per angle section
                 
        # append pressures for all angle data
        for i,a in enumerate(angles_dict):
            angles_dict[a].append(growth_angular_sections[i])
    
    ##### just for testing angular sections
    # plt.figure()    # check the segmentation
    # plt.imshow(mask_t0)
    # plt.scatter(cx_t0, cy_t0)
    # plt.show()      
    # plt.figure()    # check the angular section  
    # alpha= -70
    # mask2 = (angle >= (alpha-5)*np.pi/180.) & (angle < (alpha+5)*np.pi/180.)
    # plt.imshow(mask2)
    # plt.scatter(cx_t0, cy_t0)
    # plt.show()
    
    df = pd.DataFrame.from_dict(results)
    df.columns = ['total_growth',
                  ]
    if outfile is not None:
        df.to_excel(outfile+'_total_growth.xlsx')
    else:
        df.to_excel(folder+'//total_growth.xlsx')
     
    # save pressures for all angles
    an = pd.DataFrame.from_dict(angles_dict)
    an.columns = [a for a in range(-180, 180, 5)]
    if outfile is not None:
        an.to_excel(outfile[:-5]+'_angular_growth.xlsx')
    else:
        an.to_excel(folder+'//angular_growth.xlsx')    
    
    return



# Evaluate Angle dependet Pressures
def plot_growth(folder,  output, n_max=None,file_angulargrowth=None, file_totalgrowth=None, fontsize=7, dt=None, angle_legend=True):
    """
    Evaluate the angular growth over time for an spheroid (needs folder with PIV analysis in which also results from 'eval_angular_growth' are saved) and stores 
    the results in the output folder.
    
    - N_max may define the last image to evaluate 
    - Evaluated excel sheet with angular data is automatically searched in folder, if name differs from 'total_growth.xlsx' and 'angular_growth.xlsx' the
    user defined names can be specified using 'file_angulargrowth' or 'file_totalgrowth'
    - dt is time between consecutive images in seconds; If given a timestamp will be displayed 
    - angle_legend will draw a legend with angles if active 
    """
    from datetime import timedelta
    
   
    # read in total growth 
    if file_totalgrowth is None:   # by default looks in the PIV folder for default names
        total_growth  = pd.read_excel(folder + '//total_growth.xlsx')
    else:
        total_growth  = pd.read_excel(file_totalgrowth)  ## if stored somewhere else user can define the path
    # read in angular growth
    if file_angulargrowth is None:
        angular_growth  = pd.read_excel(folder + '//angular_growth.xlsx')
    else:
        angular_growth  = pd.read_excel(file_angulargrowth)
                
    # read in plots in folder
    plots = glob(folder+'//plot*.png')
    # make folder if not existing
    if not os.path.exists(output):
        os.makedirs(output)
    # list of evaluated angles
    angle_list = [a for a in range(-180, 180, 5)]
    # make figure
    f, (ax1, ax2) = plt.subplots(1, 2, figsize=(9,5))
            
    #second plot with only the left plot
    fb,axb = plt.subplots( figsize=(3,3))  
    # for color norm
    max_a = np.nanmax([np.nanmax(np.array(angular_growth[z])[:n_max]) for z in angle_list])  
    plt.style.use('dark_background')
    # make result dictionary


    # loop over all timesteps
    # -1 as the deformation list is one entry less (between images) compared to
    # segmentation list
    for t in tqdm(range(len(angular_growth[0][:n_max])-1)):    


              
        ang_g = []
        # go through different angles 
        for z in angle_list:        
             # append for each angle the pressure of the current timestep
             ang_g.append(np.array(angular_growth[z])[t])    
                
        # norm colors for each image individually    viridis/RdBu      
        colors = matplotlib.cm.get_cmap('viridis')(ang_g / max_a ) 
        colorcycle = 'darkred'
    
        # do some statistics
        mean = np.nanmean(ang_g)
        sd = np.nanstd(ang_g)
        CoV = np.round(sd/mean, 2)

        if angle_legend == True:
				# position of label text
            shift = 0.9
            posx_legend = [np.cos(i*np.pi/180) * shift for i in angle_list] 
            posy_legend = [np.sin(i*np.pi/180) * shift for i in angle_list] 
				# plot text
            [ax1.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
            [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
						
 			
		# now use 5 (500% growth) as maximum
        x = [np.cos(i*np.pi/180) for i in angle_list] * (ang_g / np.array([5] *len(ang_g)) ) 
        y = [np.sin(i*np.pi/180) for i in angle_list] * (ang_g / np.array([5] *len(ang_g)) ) 
				# combined figure
				# plot circles
        circle_1 = plt.Circle((0, 0), 1/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_2 = plt.Circle((0, 0), 2/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_3 = plt.Circle((0, 0), 3/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_4 = plt.Circle((0, 0), 4/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        ax1.text(0.5, 0.57, '1', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.text(0.5, 0.67, '2', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.text(0.5, 0.77, '3', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.text(0.5, 0.87, '4', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.add_artist(circle_1)
        ax1.add_artist(circle_2)
        ax1.add_artist(circle_3)
        ax1.add_artist(circle_4)
        ax1.axis('equal')  
        ax1.scatter(x,y, s=22, c=colors)
        ax1.set_xlim([-1, 1])
        ax1.set_ylim([-1, 1])
        ax1.axis('off')
		

        # annotate global growth    
        ax1.text(0.01, 0.93,'Total Growth: '+str(np.round(total_growth["total_growth"][t],2)), 
				 horizontalalignment='center',
				 verticalalignment='center',
				 transform = ax1.transAxes, fontsize=7)  
            
        # annotate CoV    
        ax1.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV), 
				 horizontalalignment='center',
				 verticalalignment='center',
				 transform = ax1.transAxes, fontsize=7)  
        
           
		# annotate time if timestep is given
        if dt is not None:
            ax1.text(0.03, 0.98,str(timedelta(seconds=t*dt)), 
                     horizontalalignment='center',
                     verticalalignment='center',
                     transform = ax1.transAxes, fontsize=11)  

					
		# show quiver plot on the right side
        plot = plt.imread(plots[t])
        ax2.imshow(plot)
        plt.axis('off')
        ax2.axis('off')
				# save figure
        f.savefig(output+'//plot_{}.png'.format(str(t).zfill(4)), dpi=200)
        ax1.cla()
        ax2.cla()
		# Single figure
		# plot circles
        circle_1 = plt.Circle((0, 0), 1/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_2 = plt.Circle((0, 0), 2/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_3 = plt.Circle((0, 0), 3/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_4 = plt.Circle((0, 0), 4/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        axb.text(0.5, 0.57, '1', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
        axb.text(0.5, 0.67, '2', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
        axb.text(0.5, 0.77, '3', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
        axb.text(0.5, 0.87, '4', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
        axb.add_artist(circle_1)
        axb.add_artist(circle_2)
        axb.add_artist(circle_3)
        axb.add_artist(circle_4)
        axb.axis('equal')  
        axb.scatter(x,y, s=22, c=colors)
        axb.set_xlim([-1, 1])
        axb.set_ylim([-1, 1])
        axb.axis('off')
        plt.savefig(output+'//single_{}.png'.format(str(t).zfill(4)), dpi=200)
        axb.cla()
       

    return 



# Evaluate Angle dependet Pressures
def plot_growth_and_pressure(folder,  output, n_max=None,file_angulargrowth=None, file_totalgrowth=None, small_pressure=False, file_angularpressure=None, fontsize=7, dt=None, angle_legend=False):
    """
    Evaluate the angular growth over time for an spheroid (needs folder with PIV analysis in which also results from 'eval_angular_growth' are saved) and stores 
    the results in the output folder.
    
    - N_max may define the last image to evaluate 
    - Evaluated excel sheet with angular data is automatically searched in folder, if name differs from 'total_growth.xlsx' and 'angular_growth.xlsx' the
    user defined names can be specified using 'file_angulargrowth' or 'file_totalgrowth'
    - dt is time between consecutive images in seconds; If given a timestamp will be displayed 
    - angle_legend will draw a legend with angles if active 
    """
    from datetime import timedelta
    
   
    # read in total growth 
    if file_totalgrowth is None:   # by default looks in the PIV folder for default names
        total_growth  = pd.read_excel(folder + '//total_growth.xlsx')
    else:
        total_growth  = pd.read_excel(file_totalgrowth)  ## if stored somewhere else user can define the path
    # read in angular growth
    if file_angulargrowth is None:
        angular_growth  = pd.read_excel(folder + '//angular_growth.xlsx')
    else:
        angular_growth  = pd.read_excel(file_angulargrowth)
        
    # read in angular pressure
    if file_angularpressure is None:
        angular_pressure  = pd.read_excel(folder + '//result_angles.xlsx')
    else:
        angular_pressure  = pd.read_excel(file_angularpressure)

    
    # read in plots in folder
    plots = glob(folder+'//plot*.png')
    # make folder if not existing
    if not os.path.exists(output):
        os.makedirs(output)
    # list of evaluated angles
    angle_list = [a for a in range(-180, 180, 5)]
    # make figure
    f, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(9+4.5,5))
            
    #second plot with only the left plot
    fb,axb = plt.subplots( figsize=(3-1,3-1))  
    # for color norm
    max_a = np.nanmax([np.nanmax(np.array(angular_growth[z])[:n_max]) for z in angle_list])  
    plt.style.use('dark_background')
    
    
    # read in pressures
    # for pressure color norm
    max_pr = np.nanmax([np.nanmax(np.array(angular_pressure[z])[:n_max]) for z in angle_list])  


    # loop over all timesteps
    for t in tqdm(range(len(angular_pressure[0][:n_max]))):   ### angular pressure has one entry less then angular growth since its calculated between images and segementation not

         
        ang_g = []
        # go through different angles 
        for z in angle_list:        
             # append for each angle the pressure of the current timestep
             ang_g.append(np.array(angular_growth[z])[t])    
                
        # norm colors for each image individually    viridis/RdBu      
        colors = matplotlib.cm.get_cmap('viridis')(ang_g / max_a ) 
        colorcycle = 'darkred'
    
        # do some statistics
        mean = np.nanmean(ang_g)
        sd = np.nanstd(ang_g)
        CoV = np.round(sd/mean, 2)

        if angle_legend == True:
				# position of label text
            shift = 0.9
            posx_legend = [np.cos(i*np.pi/180) * shift for i in angle_list] 
            posy_legend = [np.sin(i*np.pi/180) * shift for i in angle_list] 
				# plot text
            [ax1.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
            [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
						
 			
		# now use 5 (500% growth) as maximum
        x = [np.cos(i*np.pi/180) for i in angle_list] * (ang_g / np.array([5] *len(ang_g)) ) 
        y = [np.sin(i*np.pi/180) for i in angle_list] * (ang_g / np.array([5] *len(ang_g)) ) 
				# combined figure
				# plot circles
        circle_1 = plt.Circle((0, 0), 1/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_2 = plt.Circle((0, 0), 2/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_3 = plt.Circle((0, 0), 3/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_4 = plt.Circle((0, 0), 4/ 5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        ax1.text(0.5, 0.57, '1', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.text(0.5, 0.67, '2', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.text(0.5, 0.77, '3', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.text(0.5, 0.87, '4', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.add_artist(circle_1)
        ax1.add_artist(circle_2)
        ax1.add_artist(circle_3)
        ax1.add_artist(circle_4)
        ax1.axis('equal')  
        ax1.scatter(x,y, s=22, c=colors)
        ax1.set_xlim([-1, 1])
        ax1.set_ylim([-1, 1])
        ax1.axis('off')
		

        # annotate global growth    
        ax1.text(0.01, 0.93,'Total Growth: '+str(np.round(total_growth["total_growth"][t],2)), 
				 horizontalalignment='center',
				 verticalalignment='center',
				 transform = ax1.transAxes, fontsize=7)  
            
        # annotate CoV    
        ax1.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV), 
				 horizontalalignment='center',
				 verticalalignment='center',
				 transform = ax1.transAxes, fontsize=7)  
        
           
		# annotate time if timestep is given
        if dt is not None:
            ax1.text(0.03, 0.98,str(timedelta(seconds=t*dt)), 
                     horizontalalignment='center',
                     verticalalignment='center',
                     transform = ax1.transAxes, fontsize=11)  

					
		### now plot presures as second plot
        pressures = []
        # go through different angles 
        for z in angle_list:        
             # append for each angle the pressure of the current timestep
             pressures.append(np.array(angular_pressure[z])[t])    
                
        # norm colors for each image individually    viridis/RdBu      
        colors = matplotlib.cm.get_cmap('viridis')(pressures / max_pr ) 
        colorcycle = 'darkred'
    
        # do some statistics
        mean_pr =np.nanmean(pressures)
        sd_pr = np.nanstd(pressures)
        CoV_pr = np.round(sd_pr/mean_pr, 2)
        # plot pressure
        if angle_legend == True:
            # position of label text
            shift = 0.9
            posx_legend = [np.cos(i*np.pi/180) * shift for i in angle_list] 
            posy_legend = [np.sin(i*np.pi/180) * shift for i in angle_list] 
            # plot text
            [ax1.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
            [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
            [ax2.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
            [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
		
        
        # Mode that shows and norms up to 1000 Pa
        if small_pressure == False:

            # now use 1000 as maximum
            x = [np.cos(i*np.pi/180) for i in angle_list] * (pressures / np.array([1000]*len(pressures)) ) 
            y = [np.sin(i*np.pi/180) for i in angle_list] * (pressures / np.array([1000] *len(pressures)) ) 
            # combined figure
            # plot circles
            circle_100pa = plt.Circle((0, 0), 100/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            circle_250pa = plt.Circle((0, 0), 250/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            circle_500pa = plt.Circle((0, 0), 500/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            circle_750pa = plt.Circle((0, 0), 750/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            ax2.text(0.5, 0.57, '100 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.text(0.5, 0.65, '250 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.text(0.5, 0.77, '500 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.text(0.5, 0.9, '750 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.add_artist(circle_100pa)
            ax2.add_artist(circle_250pa)
            ax2.add_artist(circle_500pa)
            ax2.add_artist(circle_750pa)
            ax2.axis('equal')  
            ax2.scatter(x,y, s=22, c=colors)
            ax2.set_xlim([-1, 1])
            ax2.set_ylim([-1, 1])
            ax2.axis('off')
            # annotate CoV    
            ax2.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV_pr), 
                     horizontalalignment='center',
                     verticalalignment='center',
                     transform = ax2.transAxes, fontsize=7)  

        # mode that shows and norms up to 100 Pa
        if small_pressure == True:
		# now use 100 as maximum
              x = [np.cos(i*np.pi/180) for i in angle_list] * (pressures / np.array([100]*len(pressures)) ) 
              y = [np.sin(i*np.pi/180) for i in angle_list] * (pressures / np.array([100] *len(pressures)) ) 
        		# combined figure
        		# plot circles
              circle_a = plt.Circle((0, 0), 10/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
              circle_b = plt.Circle((0, 0), 25/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
              circle_c = plt.Circle((0, 0), 50/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
              ax2.text(0.5, 0.57, '10 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
              ax2.text(0.5, 0.65, '25 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
              ax2.text(0.5, 0.77, '50 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
              ax2.add_artist(circle_a)
              ax2.add_artist(circle_b)
              ax2.add_artist(circle_c)
              ax2.axis('equal')  
              ax2.scatter(x,y, s=22, c=colors)
              ax2.set_xlim([-1, 1])
              ax2.set_ylim([-1, 1])
              ax2.axis('off')
		      # annotate CoV    
              ax2.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV), 
                       horizontalalignment='center',
                           verticalalignment='center',
                          transform = ax2.transAxes, fontsize=9.2)  

        
        
        
        # show quiver plot on the right side
        plot = plt.imread(plots[t])
        ax3.imshow(plot)
        plt.axis('off')
        ax3.axis('off')
				# save figure
        f.savefig(output+'//plot_{}.png'.format(str(t).zfill(4)), dpi=200)
        ax1.cla()
        ax2.cla()
        ax3.cla()

    return 



def eval_angular_shape(folder, outfile=None, scale=None):
    """
    Evaluates the shape in 5° angular sections (parelell to the agular pressure and angular growth analyis)  of the 
    spheroid using the continuously tracked segmentation (therfore the "continous_segmentation" must be set to True
    in the function "compute_displacement_series"). In each angle section the maximal distance towards center is calculated 
    per timestep. 
    
    - scale: if given the resulting distance values will converted. If not analysis gives pixelvalues
    - folder: contains PIV results with segmentation
    - outfile: can be used to change the standard name of the excel file - if none excel files are created 
    in the given folder
    - scale converts the 
    
    Returns:
    saves an excel file with the organoid shape (distance to (initial) spheroid center per angle shell) 
    """
       
    ## load segmentations
    seg_files = natsorted(glob(folder+'/seg*.npy'))
       
    # initial spheroid mask 
    mask_t0 = load(seg_files[0])['mask']
    # intitial spheoir center for all further analysis
    cx_t0, cy_t0 = load(seg_files[0])['centroid']
    
    # Calculate Angle + Distances 
    y,x = np.indices(mask_t0.shape)
    dx = x - cx_t0
    dy = y - cy_t0
    distance = np.sqrt(dx ** 2 + dy ** 2)  # dist to center
    if scale:   ### convert units  - if none pixel values
        distance = distance * scale
          
    angle = np.arctan2(dy, dx) 
           
    # create dict with all angles
    angles_dict = {str(a): [] for a in range(-180, 180, 5)}
    
      
    # loop over series of PIV segmentations
    for seg_file in tqdm(seg_files):
        # load current segmentation
        mask = load(seg_file)['mask']      
             
        # shape in angular section
        angular_sections = []
        distance_angular_sections = []
        for alpha in range(-180, 180, 5):
            ### look only in this angle shell and in the segmented mask
            mask2 = (angle >= (alpha-5)*np.pi/180.) & (angle < (alpha+5)*np.pi/180.) & (mask==1)  
                   
            angular_sections.append(alpha)

            try:
                distance_angular_sections.append(np.nanmax(distance[mask2]))  #### growth per angle section
            except:
                distance_angular_sections.append(np.nan)
                
        # append pressures for all angle data
        for i,a in enumerate(angles_dict):
            angles_dict[a].append(distance_angular_sections[i])
         
    # save pressures for all angles
    an = pd.DataFrame.from_dict(angles_dict)
    an.columns = [a for a in range(-180, 180, 5)]
    if outfile is not None:
        an.to_excel(outfile[:-5]+'_angular_shape.xlsx')
    else:
        an.to_excel(folder+'//angular_shape.xlsx')    
    
    return

# Evaluate Angle dependet Pressures
def plot_shape_and_pressure(folder,  output, n_max=None,file_angularshape=None, small_pressure=False, file_angularpressure=None, fontsize=7, dt=None, angle_legend=False):
    """
    Evaluate the angular shape over time for an spheroid (needs folder with PIV analysis in which also results from 'eval_angular_shape' are saved) and stores 
    the results in the output folder.
    
    - N_max may define the last image to evaluate 
    - Evaluated excel sheet with angular data is automatically searched in folder, if name differs from 'total_growth.xlsx' and 'angular_growth.xlsx' the
    user defined names can be specified using 'file_angulargrowth' or 'file_totalgrowth'
    - dt is time between consecutive images in seconds; If given a timestamp will be displayed 
    - angle_legend will draw a legend with angles if active 
    """
    from datetime import timedelta
    
   
    
    # read in angular growth
    if file_angularshape is None:
        angular_shape  = pd.read_excel(folder + '//angular_shape.xlsx')
    else:
        angular_shape  = pd.read_excel(file_angularshape)
        
    # read in angular pressure
    if file_angularpressure is None:
        angular_pressure  = pd.read_excel(folder + '//result_angles.xlsx')
    else:
        angular_pressure  = pd.read_excel(file_angularpressure)

    
    # read in plots in folder
    plots = glob(folder+'//plot*.png')
    # make folder if not existing
    if not os.path.exists(output):
        os.makedirs(output)
    # list of evaluated angles
    angle_list = [a for a in range(-180, 180, 5)]
    # make figure
    f, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(9+4.5,5))
            
    #second plot with only the left plot
    fb,axb = plt.subplots( figsize=(3-1,3-1))  
    # for color norm
    max_a = np.nanmax([np.nanmax(np.array(angular_shape[z])[:n_max]) for z in angle_list])  
    plt.style.use('dark_background')
    
    
    # read in pressures
    # for pressure color norm
    max_pr = np.nanmax([np.nanmax(np.array(angular_pressure[z])[:n_max]) for z in angle_list])  


    # loop over all timesteps
    for t in tqdm(range(len(angular_pressure[0][:n_max]))):   ### angular pressure has one entry less then angular growth since its calculated between images and segementation not

         
        ang_g = []
        # go through different angles 
        for z in angle_list:        
             # append for each angle the pressure of the current timestep
             ang_g.append(np.array(angular_shape[z])[t])    
                
        # norm colors for each image individually    viridis/RdBu      
        colors = matplotlib.cm.get_cmap('viridis')(ang_g / max_a ) 
        colorcycle = 'darkred'
    
        # do some statistics
        mean = np.nanmean(ang_g)
        sd = np.nanstd(ang_g)
        CoV = np.round(sd/mean, 2)

        if angle_legend == True:
				# position of label text
            shift = 0.9
            posx_legend = [np.cos(i*np.pi/180) * shift for i in angle_list] 
            posy_legend = [np.sin(i*np.pi/180) * shift for i in angle_list] 
				# plot text
            [ax1.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
            [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
						
 			
		# now use 5 (500% growth) as maximum
        x = [np.cos(i*np.pi/180) for i in angle_list] * (ang_g/ np.array([max_a] *len(ang_g)) ) 
        y = [np.sin(i*np.pi/180) for i in angle_list] * (ang_g/ np.array([max_a] *len(ang_g)) ) 

        
        circle_1 = plt.Circle((0, 0), 0.25, color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_2 = plt.Circle((0, 0), 0.5 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_3 = plt.Circle((0, 0), 0.75 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        circle_4 = plt.Circle((0, 0), 1 , color=colorcycle, zorder=10, fill=False, linestyle='--')
        # ax1.text(0.5,  0.25, f"{np.round(max_a*0.25)}", color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        # ax1.text(0.5, 0.5,f"{np.round(max_a*0.5)}", color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        # ax1.text(0.5,  0.75, f"{np.round(max_a*0.75)}", color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.text(0.5,  1, f"{np.round(max_a)}", color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
        ax1.add_artist(circle_1)
        ax1.add_artist(circle_2)
        ax1.add_artist(circle_3)
        ax1.add_artist(circle_4)
        ax1.axis('equal')  
        ax1.scatter(x,y, s=22, c=colors)
        ax1.set_xlim([-1.5, 1.5])
        ax1.set_ylim([-1.5, 1.5])
        ax1.axis('off')
		
            
        # annotate CoV    
        ax1.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV), 
				 horizontalalignment='center',
				 verticalalignment='center',
				 transform = ax1.transAxes, fontsize=7)  
        
           
		# annotate time if timestep is given
        if dt is not None:
            ax1.text(0.03, 0.98,str(timedelta(seconds=t*dt)), 
                     horizontalalignment='center',
                     verticalalignment='center',
                     transform = ax1.transAxes, fontsize=11)  

					
		### now plot presures as second plot
        pressures = []
        # go through different angles 
        for z in angle_list:        
             # append for each angle the pressure of the current timestep
             pressures.append(np.array(angular_pressure[z])[t])    
                
        # norm colors for each image individually    viridis/RdBu      
        colors = matplotlib.cm.get_cmap('viridis')(pressures / max_pr ) 
        colorcycle = 'darkred'
    
        # do some statistics
        mean_pr =np.nanmean(pressures)
        sd_pr = np.nanstd(pressures)
        CoV_pr = np.round(sd_pr/mean_pr, 2)
        # plot pressure
        if angle_legend == True:
            # position of label text
            shift = 0.9
            posx_legend = [np.cos(i*np.pi/180) * shift for i in angle_list] 
            posy_legend = [np.sin(i*np.pi/180) * shift for i in angle_list] 
            # plot text
            [ax1.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
            [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
            [ax2.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
            [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
		
        
        # Mode that shows and norms up to 1000 Pa
        if small_pressure == False:

            # now use 1000 as maximum
            x = [np.cos(i*np.pi/180) for i in angle_list] * (pressures / np.array([1000]*len(pressures)) ) 
            y = [np.sin(i*np.pi/180) for i in angle_list] * (pressures / np.array([1000] *len(pressures)) ) 
            # combined figure
            # plot circles
            circle_100pa = plt.Circle((0, 0), 100/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            circle_250pa = plt.Circle((0, 0), 250/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            circle_500pa = plt.Circle((0, 0), 500/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            circle_750pa = plt.Circle((0, 0), 750/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
            ax2.text(0.5, 0.57, '100 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.text(0.5, 0.65, '250 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.text(0.5, 0.77, '500 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.text(0.5, 0.9, '750 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
            ax2.add_artist(circle_100pa)
            ax2.add_artist(circle_250pa)
            ax2.add_artist(circle_500pa)
            ax2.add_artist(circle_750pa)
            ax2.axis('equal')  
            ax2.scatter(x,y, s=22, c=colors)
            ax2.set_xlim([-1, 1])
            ax2.set_ylim([-1, 1])
            ax2.axis('off')
            # annotate CoV    
            ax2.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV_pr), 
                     horizontalalignment='center',
                     verticalalignment='center',
                     transform = ax2.transAxes, fontsize=7)  

        # mode that shows and norms up to 100 Pa
        if small_pressure == True:
		# now use 100 as maximum
              x = [np.cos(i*np.pi/180) for i in angle_list] * (pressures / np.array([100]*len(pressures)) ) 
              y = [np.sin(i*np.pi/180) for i in angle_list] * (pressures / np.array([100] *len(pressures)) ) 
        		# combined figure
        		# plot circles
              circle_a = plt.Circle((0, 0), 10/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
              circle_b = plt.Circle((0, 0), 25/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
              circle_c = plt.Circle((0, 0), 50/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
              ax2.text(0.5, 0.57, '10 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
              ax2.text(0.5, 0.65, '25 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
              ax2.text(0.5, 0.77, '50 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax2.transAxes,fontsize=fontsize)
              ax2.add_artist(circle_a)
              ax2.add_artist(circle_b)
              ax2.add_artist(circle_c)
              ax2.axis('equal')  
              ax2.scatter(x,y, s=22, c=colors)
              ax2.set_xlim([-1, 1])
              ax2.set_ylim([-1, 1])
              ax2.axis('off')
		      # annotate CoV    
              ax2.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV), 
                       horizontalalignment='center',
                           verticalalignment='center',
                          transform = ax2.transAxes, fontsize=9.2)  

        
        
        
        # show quiver plot on the right side
        plot = plt.imread(plots[t])
        ax3.imshow(plot)
        plt.axis('off')
        ax3.axis('off')
				# save figure
        f.savefig(output+'//plot_{}.png'.format(str(t).zfill(4)), dpi=200)
        ax1.cla()
        ax2.cla()
        ax3.cla()

    return 




import os
import dill
import shutil
import numpy as np
from glob import glob
from subprocess import call
from subprocess import Popen
from subprocess import PIPE
from time import sleep
from tqdm import tqdm
from scipy.interpolate import LinearNDInterpolator
from saenopy import Solver
import saenopy
from saenopy.materials import SemiAffineFiberMaterial
import pandas as pd
from pathlib import Path
try: 
    # if new saenopy format use this function
    from saenopy.solver import load_solver
except:
    # if not rename the load function from the old saenopy version
    import saenopy.load as load_solver

def read_meshfile(meshfile, r_inner=None, r_outer=None):
    # open mesh file
    with open(meshfile, 'r') as f:
        lines = f.readlines()

    # if r_inner or r_outer are defined in the mesh-file, ignore user input
    meshinfo = {}
    try:
        paragraph = lines[lines.index('$Jointforces\n') + 1:lines.index('$EndJointforces\n')]
        for line in paragraph:
            key, value = line.strip().split('=')
            try:
                meshinfo[key] = float(value)
            except ValueError:
                meshinfo[key] = value

        r_inner = meshinfo['r_inner']
        r_outer = meshinfo['r_outer']

        print('Geometry for spherical contraction is defined in mesh file (r_inner={:.2f}, r_outer={:.2f}).'.format(r_inner, r_outer))
        if (r_inner is not None) or (r_outer is not None):
            print('Will ignore user-defined values of r_inner and r_outer.')
    except ValueError:
        if r_inner is None:
            raise ValueError('r_inner not defined')
        if r_outer is None:
            raise ValueError('r_outer not defined')
            
    # scale radii to meter
    r_inner *= 10**-6
    r_outer *= 10**-6
    
    # transform nodes and connection in SAENO format
    # nodes
    index_nodes = lines.index('$Nodes\n')
    n_nodes = int(lines[index_nodes + 1])

    coords = np.zeros((n_nodes, 3))
    for i in range(n_nodes):
        coords[i] = np.array([float(x) for x in lines[i + index_nodes + 2].split()[1:]])
        
      
    # connections
    index_elements = lines.index('$Elements\n')
    n_elements = int(lines[index_elements + 1])

    tets = np.zeros((n_elements, 4))
    for i in range(n_elements):
        tets[i] = lines[i + index_elements + 2].split()[-4:]

    # to start with 0 and not 1
    tets -= 1
    
    return coords, tets, r_inner, r_outer
    

def spherical_contraction_solver(meshfile, outfolder, pressure, material, r_inner=None, r_outer=None, logfile=False, initial_displacenemts=None, max_iter = 600, step = 0.0033, conv_crit = 0.01):
    
    
    coords, tets, r_inner, r_outer = read_meshfile(meshfile, r_inner, r_outer)

    # read in material parameters
    K_0 = material['K_0']
    D_0 = material['D_0']
    L_S = material['L_S']
    D_S = material['D_S']

    # create output folder if it does not exist, print warning otherwise
    if not os.path.exists(outfolder):
        os.makedirs(outfolder)
    else:
        print('WARNING: Output folder already exists! ({})'.format(outfolder))


    # Initialize saenopy solver opbject
    M = Solver()
    material_saenopy = SemiAffineFiberMaterial(K_0, D_0, L_S, D_S)
    M.set_material_model(material_saenopy)
    M.set_nodes(coords)
    M.set_tetrahedra(tets)
     
    # define boundary conditions
    distance = np.sqrt(np.sum(coords ** 2., axis=1))
    mask_inner = distance < r_inner * 1.001
    mask_outer = distance > r_outer * 0.999

    # Save Node Density at inner and outer sphere
    # Area per inner node
    A_node_inner = (np.pi*4*(r_inner)**2)/np.sum(mask_inner) 
    # simple sqrt as spacing
    inner_spacing = np.sqrt(A_node_inner)    
    
    # Area per outer node
    A_node_outer = (np.pi*4*(r_outer)**2)/np.sum(mask_outer)   
    # simple sqrt as spacing
    outer_spacing = np.sqrt(A_node_outer)
    
    print ('Inner node spacing: '+str(inner_spacing*1e6)+'µm')
    print ('Outer node spacing: '+str(outer_spacing*1e6)+'µm')

    
    # displacements are everywhere NaN
    bcond_displacement = np.zeros((len(coords), 3))*np.nan
    # except of a the outer border
    bcond_displacement[mask_outer] = 0
    
    # forces are everywhere 0
    bcond_forces = np.zeros((len(coords), 3))
    # except at the outer boder there they are NaN
    bcond_forces[mask_outer] = np.nan
    # and at the innter border they depend on the pressure
    bcond_forces[mask_inner] = coords[mask_inner]
    bcond_forces[mask_inner] /= distance[mask_inner, None]
    A_inner = 4 * np.pi * r_inner ** 2.
    force_per_node = pressure * A_inner / np.sum(mask_inner)
    bcond_forces[mask_inner, :3] *= force_per_node
    
    # give the boundary conditions to the solver
    M.set_boundary_condition(bcond_displacement, bcond_forces)

    if initial_displacenemts is not None:
        M.set_initial_displacements(initial_displacenemts)

    # create info file with all relevant parameters of the simulation
    parameters = r"""K_0 = {}
D_0 = {}
L_S = {}
D_S = {}
PRESSURE = {}
FORCE_PER_SURFACE_NODE = {}
INNER_RADIUS = {} µm
OUTER_RADIUS = {} µm
INNER_NODE_SPACING = {} µm
OUTER_NODE_SPACING = {} µm
SURFACE_NODES = {}
TOTAL_NODES = {}""".format(K_0, D_0, L_S, D_S, pressure, force_per_node, r_inner*1e6, r_outer*1e6, inner_spacing*1e6,
                           outer_spacing*1e6, np.sum(mask_inner), len(coords))

    with open(outfolder + "/parameters.txt", "w") as f:
        f.write(parameters)
        
    # solve the boundary problem
    M.solve_boundarycondition(step_size=step, max_iterations=max_iter, rel_conv_crit=conv_crit, relrecname=outfolder + "/relrec.txt") #, verbose=True
    M.save(outfolder + "/solver.saenopy")
    
  
def distribute_solver(func, const_args, var_arg='pressure', start=0.1, end=1000, n=120, log_scaling=True, n_cores=None, get_initial=True, max_iter = 600, step = 0.0033, conv_crit = 0.01, callback=None):
    # get_intial = True takes the deformationfield from previous simulation as start values for the next simulations, which reduces computation time
    
    # by default use spherical contraction as function
    func = spherical_contraction_solver
    
    # set values if we have constant arguments here instead of default
    if "max_iter" in const_args:
        max_iter = const_args["max_iter"]    
    if "step" in const_args:
        step = const_args["step"]   
    if "conv_crit" in const_args:
        conv_crit = const_args["conv_crit"] 
        
  
    if n_cores is None:
        n_cores = os.cpu_count()

    if log_scaling:
        values = np.logspace(np.log10(start), np.log10(end), n, endpoint=True)
    else:
        values = np.linspace(np.log10(start), np.log10(end), n, endpoint=True)

    outfolder = const_args['outfolder']
    del const_args['outfolder']

    if not os.path.exists(outfolder):
        os.makedirs(outfolder)

    np.savetxt(outfolder+'/'+var_arg+'-values.txt', values)

    values = list(values)

    if n_cores == 1:
        U = None
        for index, v in enumerate(values):
            func(const_args["meshfile"], outfolder + '/simulation' + str(index).zfill(6),
                    v, const_args["material"], None, None, False,  U, max_iter, step, conv_crit)
            # get the last displacement
            name = outfolder + '/simulation' + str(index).zfill(6) + "/solver.npz"
            if os.path.exists(name):
                M = load_solver(name)
                U = M.U
            # optionally call the callback
            if callback is not None:
                callback(index, len(values))
    else:
        index = 0
        processes = []
        import multiprocessing
        while True:
            processes = [p for p in processes if p.is_alive()]

            if len(processes) < n_cores and index < len(values):
                v = values[index]
                U = None
                if get_initial==True:
                    for i in range(index-3, -1, -1):
                        name = outfolder+'/simulation'+str(i).zfill(6) + "/solver.npz"
                        if os.path.exists(name):
                            M = load_solver(name)
                            U = M.U
                            break
                p = multiprocessing.Process(target=func,
                                            args=(const_args["meshfile"],  outfolder+'/simulation'+str(index).zfill(6),
                                                  v, const_args["material"], None, None, False, U,  max_iter, step, conv_crit))

                p.start()
                processes.append(p)
                if callback is not None:
                    callback(index, len(values))
                index += 1
            sleep(1.)

            if len(processes) == 0:
                break
    if callback is not None:
        callback(index, len(values))
    return
    
# only needed for old saenopy files  
def saenopy_fake_load(c):
    Solver = saenopy.Solver()
    A = np.load(c,allow_pickle=True)
    Solver.mesh.forces =  A["f"]  
    #Solver.mesh.regularisation_mask = A["reg_mask"] 
    Solver.mesh.nodes = A["R"]  
    Solver.mesh.displacements = A["U"]  
    Solver.get_contractility()
    return Solver


def extract_deformation_curve_solver(folder, x):
    # get simulation parameters
    with open(folder+'/parameters.txt', 'r') as f:
        lines = f.readlines()

    parameters = {}
    for line in lines:
        try:
            key, value = line.split('= ')
            value = value.split(' ')[0]
            parameters[key.strip()] = float(value.strip())
        except:
            pass

    # load data
    try:
        M = load_solver(folder + "/solver.npz")
        coords = M.mesh.nodes 
        displ  = M.mesh.displacements
    # Support older saenopy versions
    except:    
        try:   
            M = load_solver(folder + "/solver.npz")
            coords = M.R 
            displ = M.U
        except: #
            M = saenopy_fake_load(folder + "/solver.npz")
            coords = M.mesh.nodes 
            displ = M.mesh.displacements
       
    
    
    # compute binned normed displacements and normed coordinates
    u = np.sqrt(np.sum(coords ** 2., axis=1)) / (parameters['INNER_RADIUS']*10**-6)
    v = np.sqrt(np.sum(displ ** 2., axis=1)) / (parameters['INNER_RADIUS']*10**-6)

    y = np.array([np.nanmedian(v[(u >= x[i]) & (u < x[i + 1])]) for i in range(len(x) - 1)])

    # save results
    results = {'pressure': parameters['PRESSURE'], 'bins': x, 'displacements': y}
    return results
  
    
def create_lookup_table_solver(folder, x0=1, x1=50, n=100):
    subfolders = glob(folder+'/*/')

    x = np.logspace(np.log10(x0), np.log10(x1), n+1, endpoint=True)
    x_center = 10**(0.5*(np.log10(x[1:]) + np.log10(x[:-1])))

    pressure_values = []
    displacement_curves = []

    for subfolder in tqdm(subfolders):
        res = extract_deformation_curve_solver(subfolder, x)
        pressure_values.append(res['pressure'])
        displacement_curves.append(res['displacements'])

    pressure_values = np.array(pressure_values)
    displacement_curves = np.array(displacement_curves)

    return {'pressure': pressure_values, 'x': x_center, 'y': displacement_curves}
  

  
    
"""
Command_line_version 
"""   
    
    
    
    
def spherical_contraction(meshfile, outfolder, pressure, material, r_inner=None, r_outer=None, logfile=False,  max_iter = 600, step = 0.0033, conv_crit = 0.01):
    # open mesh file
    with open(meshfile, 'r') as f:
        lines = f.readlines()

    # if r_inner or r_outer are defined in the mesh-file, ignore user input
    meshinfo = {}
    try:
        paragraph = lines[lines.index('$Jointforces\n') + 1:lines.index('$EndJointforces\n')]
        for line in paragraph:
            key, value = line.strip().split('=')
            try:
                meshinfo[key] = float(value)
            except ValueError:
                meshinfo[key] = value

        r_inner = meshinfo['r_inner']
        r_outer = meshinfo['r_outer']

        print('Geometry for spherical contraction is defined in mesh file (r_inner={:.2f}, r_outer={:.2f}).'.format(r_inner, r_outer))
        if (r_inner is not None) or (r_outer is not None):
            print('Will ignore user-defined values of r_inner and r_outer.')
    except ValueError:
        if r_inner is None:
            raise ValueError('r_inner not defined')
        if r_outer is None:
            raise ValueError('r_outer not defined')
    
    # scale radii to meter
    r_inner *= 10**-6
    r_outer *= 10**-6

    # read in material parameters
    K_0 = material['K_0']
    D_0 = material['D_0']
    L_S = material['L_S']
    D_S = material['D_S']

    # create output folder if it does not exist, print warning otherwise
    if not os.path.exists(outfolder):
        os.makedirs(outfolder)
    else:
        print('WARNING: Output folder already exists! ({})'.format(outfolder))

    # transform nodes and connection in SAENO format
    # nodes
    index_nodes = lines.index('$Nodes\n')
    n_nodes = int(lines[index_nodes + 1])

    coords = np.zeros((n_nodes, 3))
    for i in range(n_nodes):
        coords[i] = np.array([np.float(x) for x in lines[i + index_nodes + 2].split()[1:]])
    np.savetxt(outfolder + '/coords.dat', coords)

    # connections
    index_elements = lines.index('$Elements\n')
    n_elements = int(lines[index_elements + 1])

    tets = np.zeros((n_elements, 4))
    for i in range(n_elements):
        tets[i] = lines[i + index_elements + 2].split()[-4:]
    np.savetxt(outfolder + '/tets.dat', tets, fmt='%i')

    # define boundary conditions
    distance = np.sqrt(np.sum(coords ** 2., axis=1))
    mask_inner = distance < r_inner * 1.001
    mask_outer = distance > r_outer * 0.999

    # Save Node Density at inner and outer sphere
    # Area per inner node
    A_node_inner = (np.pi*4*(r_inner)**2)/np.sum(mask_inner) 
    # simple sqrt as spacing
    inner_spacing = np.sqrt(A_node_inner)    
    
    # Area per outer node
    A_node_outer = (np.pi*4*(r_outer)**2)/np.sum(mask_outer)   
    # simple sqrt as spacing
    outer_spacing = np.sqrt(A_node_outer)
    
    print ('Inner node spacing: '+str(inner_spacing*1e6)+'µm')
    print ('Outer node spacing: '+str(outer_spacing*1e6)+'µm')

    bcond = np.zeros((len(coords), 4))
    bcond[:, 3] = 1.

    # fixed displacements for outer boundary
    bcond[mask_outer, 3] = 0

    # fixed non-zero force at spheroid surface
    bcond[mask_inner, :3] = coords[mask_inner, :3]
    bcond[mask_inner, :3] /= distance[mask_inner, None]

    A_inner = 4 * np.pi * r_inner ** 2.
    force_per_node = pressure * A_inner / np.sum(mask_inner)
    bcond[mask_inner, :3] *= force_per_node

    np.savetxt(outfolder + '/bcond.dat', bcond)

    # define initial configuration
    iconf = np.zeros((len(coords), 3))
    np.savetxt(outfolder + '/iconf.dat', iconf)

    # create config file for SAENO    
    config = r"""MODE = relaxation
BOXMESH = 0
FIBERPATTERNMATCHING = 0
REL_CONV_CRIT = {}
REL_ITERATIONS = {}
REL_SOLVER_STEP = {}
K_0 = {}
D_0 = {}
L_S = {}
D_S = {}
CONFIG = {}\config.txt
DATAOUT = {}""".format(conv_crit, max_iter, step, K_0, D_0, L_S, D_S, os.path.abspath(outfolder), os.path.abspath(outfolder))

    with open(outfolder + "/config.txt", "w") as f:
        f.write(config)

    # create info file with all relevant parameters of the simulation
    parameters = r"""K_0 = {}
D_0 = {}
L_S = {}
D_S = {}
PRESSURE = {}
FORCE_PER_SURFACE_NODE = {}
INNER_RADIUS = {} µm
OUTER_RADIUS = {} µm
INNER_NODE_SPACING = {} µm
OUTER_NODE_SPACING = {} µm
SURFACE_NODES = {}
TOTAL_NODES = {}""".format(K_0, D_0, L_S, D_S, pressure, force_per_node, r_inner*1e6, r_outer*1e6, inner_spacing*1e6,
                           outer_spacing*1e6, np.sum(mask_inner), len(coords))

    with open(outfolder + "/parameters.txt", "w") as f:
        f.write(parameters)

    # Create log file if activated
    if logfile:
        
        # create log file with system output
        logfile = open(outfolder + "/saeno_log.txt", 'w')
        cmd = Popen(["saenopy","CONFIG","{}//config.txt".format(os.path.abspath(outfolder))], stdout=PIPE, 
                    universal_newlines=True, shell=False)
        # print and save a reduced version of saeno log
        for line in cmd.stdout:
            if not '%' in line:
                print (line, end='')
                logfile.write(str(line))
        # close again to avoid loops            
        cmd.stdout.close()            
        
    # if false just show the non reduced system output    
    else:
        cmd = call(["saenopy","CONFIG", "{}//config.txt".format(os.path.abspath(outfolder))])

    # copy result files from "*_py2" folder
    # for filename in glob(outfolder+'_py2/*.*'):
    #     shutil.copy(filename, outfolder)

    # # remove "*_py2" folder
    # shutil.rmtree(outfolder+'_py2')


def distribute(func, const_args, var_arg='pressure', start=0.1, end=1000, n=120, log_scaling=True, n_cores=None):
    if n_cores is None:
        n_cores = os.cpu_count()

    if log_scaling:
        values = np.logspace(np.log10(start), np.log10(end), n, endpoint=True)
    else:
        values = np.linspace(np.log10(start), np.log10(end), n, endpoint=True)

    outfolder = const_args['outfolder']
    del const_args['outfolder']

    if not os.path.exists(outfolder):
        os.makedirs(outfolder)

    np.savetxt(outfolder+'/'+var_arg+'-values.txt', values)

    values = list(values)
    index = list(np.arange(len(values)))
    processes = []

    
    
    while True:
        if len(processes) < n_cores and len(values) > 0:
            command = '''python -c "import jointforces; import jointforces as jf; {}('''.format(func)
            for key in const_args:
                if isinstance(const_args[key], str):
                    command += '''{}='{}','''.format(key, const_args[key])
                else:
                    command += '''{}={},'''.format(key, const_args[key])
            command += '''outfolder='{}', {}={})"'''.format(outfolder+'/simulation'+str(index[0]).zfill(6), var_arg, values[0])
            print ("Simulations in the queue: "+str(len(values)))
            processes.append(Popen(command, shell='True'))
            del values[0]
            del index[0]

        sleep(1.)

        processes = [p for p in processes if p.poll() is None]

        if len(processes) == 0:
            break


def extract_deformation_curve(folder, x):
    # get simulation parameters
    with open(folder+'/parameters.txt', 'r') as f:
        lines = f.readlines()

    parameters = {}
    for line in lines:
        try:
            key, value = line.split('= ')
            value = value.split(' ')[0]
            parameters[key.strip()] = float(value.strip())
        except:
            pass

    # load coordinates
    coords = np.genfromtxt(folder + '/R.dat')

    # load displacements
    displ = np.genfromtxt(folder + '/U.dat')

    # compute binned normed displacements and normed coordinates
    u = np.sqrt(np.sum(coords ** 2., axis=1)) / (parameters['INNER_RADIUS']*10**-6)
    v = np.sqrt(np.sum(displ ** 2., axis=1)) / (parameters['INNER_RADIUS']*10**-6)

    y = np.array([np.nanmedian(v[(u >= x[i]) & (u < x[i + 1])]) for i in range(len(x) - 1)])

    # save results
    results = {'pressure': parameters['PRESSURE'], 'bins': x, 'displacements': y}
    return results


def create_lookup_table(folder, x0=1, x1=50, n=100):
    subfolders = glob(folder+'/*/')

    x = np.logspace(np.log10(x0), np.log10(x1), n+1, endpoint=True)
    x_center = 10**(0.5*(np.log10(x[1:]) + np.log10(x[:-1])))

    pressure_values = []
    displacement_curves = []

    for subfolder in tqdm(subfolders):
        res = extract_deformation_curve(subfolder, x)
        pressure_values.append(res['pressure'])
        displacement_curves.append(res['displacements'])

    pressure_values = np.array(pressure_values)
    displacement_curves = np.array(displacement_curves)

    return {'pressure': pressure_values, 'x': x_center, 'y': displacement_curves}


def create_lookup_functions(lookup_table):
    pressure = lookup_table['pressure']
    distance = lookup_table['x']
    displacement = lookup_table['y']

    log_pressure = np.log(pressure)
    log_distance = np.log(distance)

    x, y = np.meshgrid(log_distance, log_pressure)

    mask = ~(np.isnan(x) | np.isnan(y) | np.isnan(displacement))

    x = x[mask]
    y = y[mask]
    displacement = displacement[mask]

    f = LinearNDInterpolator(np.array([x, y]).T, displacement)

    def get_displacement(distance, pressure):
        return f(np.log(distance), np.log(pressure))

    f_inv = LinearNDInterpolator(np.array([x, displacement]).T, y)

    def get_pressure(distance, displacement):
        return np.exp(f_inv(np.log(distance), displacement))

    return get_displacement, get_pressure


""" OLD SAVING FUNCITON THAT SHOULD NOT BE USED ANYMORE
def save_lookup_functions(get_displacement, get_pressure, outfile):
    with open(outfile, 'wb') as f:
        dill.dump((get_displacement, get_pressure), f)


def save_lookup_functions(get_displacement, get_pressure, outfile):
    np.save(outfile,(get_displacement, get_pressure))
          

def save_lookup_table(lookup_table, outfile):
    with open(outfile, 'wb') as f:
        dill.dump(lookup_table, f)
"""      
  
def save_lookup_table(lookup_table, outfile):
    np.save(outfile,lookup_table)

# either loads lookupfunctions or loads lookuptable and create the functions thereof
def load_lookup_functions(file):
    try:   
        lookup_table = np.load(file,allow_pickle=True).item()
        get_displacement, get_pressure = create_lookup_functions(lookup_table)
        print ("Found Lookuptable, and created lookup functions")
        
    except:  # In case of old saved formats    
        try:
            with open(file, 'rb') as f:
                get_displacement, get_pressure = dill.load(f)
            print ("Found Lookuptable in Old Format, and created lookup functions")
        except:
            with open(file, 'rb') as f:
                lookup_table = dill.load(f)
                get_displacement, get_pressure = create_lookup_functions(lookup_table)
            print ("Found Lookuptable in Old Format, and created lookup functions")    
    return get_displacement, get_pressure


""" OLD interpolation function
# def linear_lookup_interpolator(emodulus, output_newtable="new-lin-lookup.pkl", reference_folder=None):
#     
#     Create individual lookup-tables for linear materials by shifting a reference lookuptable for a linear fiber material.
    
#     For linear fiber materials the following relation is used:  k0 = 6 * E_Modulus  (for a possion ration of 0.25, see Steinwachs[2015]) 
    
#     Original simulation reached up to 10 000 Pa for a simulated k0 = 2364 (emodul ~394 Pa) - interpolation should be usefull for a wide range of
#      emoduli - however keep in mind that there might be limits for extreme Emoduli-pressures combination due to the range of the original simulations 
#      (in such a case a constant maximal or minimal-pressure value will be returned since there are no better fitting simulations)
    
#     emodulus:ArithmeticError Desired Youngs modulus for which the linear lookup table is created
#     reference_folder: Folder containing the reference lookup functions and reference interpolators to create
#     the new look up table; By default (None) will search for the reference files automatically  
#     output_newtable: name for the new reference table (needs .pkl ending)
#     
    
#     # if not specified differently we find the correct reference files automatically
#     if not reference_folder:
#         import jointforces
#         reference_folder = Path(jointforces.__file__).parent / ".." / "docs" / "data" / "linear_reference_table"
        
#     # load in  reference lookuptable to for a simulated k2364 (emodul ~394 Pa) up to 10 000 Pa 
#     get_displacement_ref, get_pressure_ref = load_lookup_functions(os.path.join(reference_folder,'linear-ref-functions-k2364.pkl'))

#     # load in in reference interpolators
#     f_ref,f_inv_ref = load_lookup_functions(os.path.join(reference_folder,'linear-interp-f-finv-k2364.pkl')) 
    
#     # shift the table accordingly 
#     def get_displacement_new(distance, pressure):
#         return f_ref(np.log(distance), np.log(pressure)) * 2364 /(emodulus*6)

#     def get_pressure_new(distance, displacement):
#         return np.exp(f_inv_ref(np.log(distance), displacement)) * (emodulus*6)/2364
    
#     # save the new lookup functions
#     save_lookup_functions(get_displacement_new, get_pressure_new, output_newtable)

#     return get_displacement_new, get_pressure_new
"""

def linear_lookup_interpolator(emodulus, output_newtable="new-lin-lookup.npy", reference_folder=None):
    """
    Create individual lookup-tables for linear materials by shifting a reference lookuptable for a linear fiber material.
    
    For linear fiber materials the following relation is used:  k0 = 6 * E_Modulus  (for a possion ration of 0.25, see Steinwachs[2015]) 
    
    Original simulation reached up to 10 000 Pa for a simulated k0 = 2364 (emodul ~394 Pa) - interpolation should be usefull for a wide range of
     emoduli - however keep in mind that there might be limits for extreme Emoduli-pressures combination due to the range of the original simulations 
     (in such a case a constant maximal or minimal-pressure value will be returned since there are no better fitting simulations)
    
    emodulus:ArithmeticError Desired Youngs modulus for which the linear lookup table is created
    reference_folder: Folder containing the reference lookup functions and reference interpolators to create
    the new look up table; By default (None) will search for the reference files automatically  
    output_newtable: name for the new reference table (needs .pkl ending)
    """
    
    # if not specified differently we find the correct reference files automatically
    if not reference_folder:
        import jointforces
        reference_folder = Path(jointforces.__file__).parent / ".." / "docs" / "data" / "linear_reference_table"
        
    # get data    
    ref_table = os.path.join(reference_folder,'linear_reference_k2364.npy')
    lookup_table = np.load(ref_table,allow_pickle=True).item()    
    pressure = lookup_table['pressure']
    distance = lookup_table['x']
    displacement = lookup_table['y']
     
    # shift table
    lookup_table_shifted = lookup_table.copy()
    lookup_table_shifted['y'] = np.array(lookup_table_shifted['y']) * 2364 /(emodulus*6) 
    save_lookup_table(lookup_table_shifted,output_newtable)

    return lookup_table_shifted



def plot_lookup_table(lookup_table, pressure=[0,10000], log_scale = True, distance=[2,50], linewidth=2, n_lines = 1000, save_plot = None,
                      fig_size=(5,4), figure=None, show=True, colorlegend = 'Pressure (Pa)'):
    """
    Create a figure of your (.pkl) material lookuptable
    
    lookup_table: path to .pkl file
    pressure: pressure range which will be plotted as [pressure_min,pressure_max] in pascal (use the range that was used to calculate the lookuptable)
    log_scale: plot logarithmically or linear if set to False
    distance: The distance to the spheroid which is plotted on the x-axis; Unit is spheroid radii
    linewidth: linewidth for the individual plots
    n_lines: number of lines plotted between the minimal and maximal pressure
    save_plot: saves the plot as png file if a path is given 
    """
    
    import matplotlib.cm as cm 
    import matplotlib.colors as colors
    import matplotlib.pyplot as plt
    
    # load lookup table
    get_displacement, get_pressure = load_lookup_functions(lookup_table)
    
    ## to avoid scaling issue for log(0) 
    if pressure[0] == 0:
        pressure[0] = 0.01     
   
    # define pressure range
    if log_scale:
        pressure_list = np.logspace(np.log(pressure[0])/np.log(10), np.log(pressure[1])/np.log(10),  num=n_lines, base=10)     
    else:
        pressure_list = np.arange(pressure[0],pressure[1], step = (pressure[1]-pressure[0])/n_lines )
      
       
      
    # make cmaps
    mycmap = cm.get_cmap('viridis')
    mynorm = colors.LogNorm(vmin=np.min(pressure_list),vmax=np.max(pressure_list))   
    c = mycmap(mynorm(pressure_list))

    # make a colorbar
    sm = plt.cm.ScalarMappable(cmap=mycmap, norm= mynorm)
    sm.set_array(c)
        
    # create distance list
    distance_list = np.arange(distance[0],distance[1], step =(distance[1]-distance[0]) / 1000 )
    
    
    # get displacements for pressures list
    displacement_list = [get_displacement(distance_list,i) for i in pressure_list]

    # create a figure if no figure is specified
    if figure is None:
        figure = plt.figure(figsize=fig_size)

    for i in range(len(displacement_list)):
        plt.plot( distance_list , displacement_list[i], c= c[i],linewidth=linewidth,alpha=0.5)
   
    # set x,y limit - go a bit above minimum and maximum
    plt.ylim(np.nanmin(displacement_list)-(0.1*np.nanmin(displacement_list)),
             np.nanmax(displacement_list)+(0.1*np.nanmax(displacement_list)))
    
    # set log if activated and go even more above minimum/maximum in log scale
    if log_scale:
        plt.loglog()
        plt.ylim(np.nanmin(displacement_list)-(0.8*np.nanmin(displacement_list)),
             np.nanmax(displacement_list)+(0.8*np.nanmax(displacement_list)))
    
    plt.grid(False) 
    plt.ylabel('Normalized deformations (u/r₀)', fontsize=10)           # plt.ylabel('Normed Deformation', fontsize=14)
    plt.xlabel('Normalized distance (r/r₀)', fontsize=10) # plt.xlabel('Normed Distance', fontsize=14)
    
    # make a colorbar
    cbar= plt.colorbar(sm, )
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel(colorlegend, rotation=270 , fontsize=10)
    plt.tight_layout()
    # save plot if specified
    if save_plot is not None:
        plt.savefig(save_plot,dpi=500)  
    # show figure if specified
    if show==True:
        plt.show()
    return figure


def plot_lookup_data(lookup_table, data_folder, timesteps=[10,30,60], distance=[2,50], linewidth=2, 
              color_line="k", color_raw="r", scatter_raw_data = True, marker_size_scatter=0.1,
              marker_size_mean=10,angle_filter=20, color_list = None,  plot_means = True,
              label_list = None, timesteps_scatter=None , result_file_name ="result.xlsx"): 
    """
    plot the pressure for certain timesteps into the lookup table as a grey line;
    scatter_raw_data option to vuisualize all deformtation-distance data at that timestep
    
    Use the function after calling "plot_lookup_table" to visualize a certain pressure within the created lookuptable
    
    lookup_table: path to .pkl filer
    data_folder: path to the evaluated folder containing result.xlsx and the dis*.npy & seg*.npy data
    timesteps: list of the timesteps to plot into the lookup function
    scatter_raw_data: option to scatter the individual deformations
    timesteps_scatter: might be used to scatter only several timesteps - if none identical to timesteps
    angle_filter: use same value as in evaluation
    if color_list & label_list is provided the given colors and labels are used when plotting the raw data
    e.g color_list=["C0","C1","C2"], label_list=["1h","3h","12h"],
    
    timesteps_scatter: might be used to scatter only several timesteps - if timesteps_scatter=None identical to timesteps;
    keep the same length as timesteps to asure same coloring :
    e.g timesteps=[1,2,3] --> timesteps_scatter= [None,2,None]  to show only second element
    """   
    import matplotlib.pyplot as plt
    from glob import glob
    from natsort import natsorted
    from .simulation import load_lookup_functions
    from .utils import load
    from .force import infer_pressure
    from tqdm import tqdm
    
    # scatter same timesteps if not spcified
    if not timesteps_scatter:
        timesteps_scatter = timesteps
    
    # read in the data
    pressure = pd.read_excel(os.path.join(data_folder,result_file_name))["Mean Pressure (Pa)"]
    
    # load lookup table
    get_displacement, get_pressure = load_lookup_functions(lookup_table)
    
    # define pressure range  # -1 since pressure list is 1 element shorter 
    # compared to image list (differences are taken)
    pressure_list = [pressure[i-1] for i in timesteps]
    print (pressure_list)
    
    # create distance list
    distance_list = np.arange(distance[0],distance[1], step =(distance[1]-distance[0]) / 1000 )

    # get displacements for pressures list
    displacement_list = [get_displacement(distance_list,i) for i in pressure_list]
   
    # draw simulations in uniform color
    for i in range(len(displacement_list)):
       plt.plot( distance_list , displacement_list[i], c= color_line, #linestyle="--",
                linewidth=linewidth,alpha=0.5,zorder=30)#, label="Simulation") 
       if i==0: # plot label once
           plt.plot( [] ,[] , c= color_line, #linestyle="--",
                    linewidth=linewidth,alpha=0.5,zorder=30, label="Simulations") 
        
    # read in segemntation
    seg_files = natsorted(glob(data_folder+'/seg*.npy'))[:np.nanmax(timesteps)]
    
    
    # load deformations
    # look for accumulated deformation files (new standard)
    d_accumulated_files = natsorted(glob(data_folder+'/def*.npy'))[:np.nanmax(timesteps)]  # do not calcualte more then the necessary time steps
    # look also for not-accummulated deformations (old standard)
    d_notaccumulated_files = natsorted(glob(data_folder+'/dis*.npy'))[:np.nanmax(timesteps)]  # do not calcualte more then the necessary time steps
    # if not-accumulated deformations are found chose different mode
    if len(d_notaccumulated_files) > len(d_accumulated_files):
        accumulated = False
        dis_files = d_notaccumulated_files
        print("Found not-accumulated deformation files (old standard) and will conduct calculations accordingly.")
    # else do the calcualtion with accumulated d eformations already
    else:
        accumulated = True
        dis_files = d_accumulated_files


    # initial spheroid radius and surface (used for force estimation)
    r0 = load(seg_files[0])['radius']
   
    u_sum = None
    v_sum = None
    
    distance_list_raw = []
    displacement_list_raw = []
    pressure_list_raw = []
    
    # loop over series of PIV results
    for (dis_file, seg_file) in tqdm(zip(dis_files, seg_files)):
        dis = load(dis_file)
        seg = load(seg_file)

        x_rav = np.ravel(dis['x'])
        y_rav = np.ravel(dis['y'])
        #print("data_points:"+str(len(x_rav)))
              
        # get deformations
        # sum up if we have not-accummulated deformations (old standard)
        if accumulated == False:
            try:
                u_sum += np.ravel(dis['u'])
                v_sum += np.ravel(dis['v'])
            except:
                u_sum = np.ravel(dis['u'])
                v_sum = np.ravel(dis['v'])
        # else read in accummulated deformations directly (new standard)
        else:
            u_sum = np.ravel(dis['u'])
            v_sum = np.ravel(dis['v'])    
            
 
        cx, cy = seg['centroid']
        distance_raw, displacement_raw, angle_raw, pressure_raw = infer_pressure(x_rav, y_rav, u_sum, v_sum, cx, cy, r0, get_pressure , angle_filter=angle_filter)
        #print (len(distance_raw)) # length of valid datapoints after angle filter      
          
        # create list with accumulated deformations 
        distance_list_raw.append(distance_raw)
        displacement_list_raw.append(displacement_raw)
        pressure_list_raw.append(pressure_raw)
        
        
    # SCATTERED RAW DATA
    #now plot the accumulated raw data at the corresponding timepoints if activated;
    # deformation and image data have a length difference of 1
    if scatter_raw_data:
        # in case colors and labels are defined; 
        if color_list and label_list:
          for ci,t in enumerate(timesteps_scatter):
            if t == None: #do not plot None elements 
                  continue
            plt.scatter(distance_list_raw[t-1],displacement_list_raw[t-1],s=marker_size_scatter,zorder=20,c = color_list[ci])      
        
            ## calculate CoV of pressure values for 72h and add to plot
            # if t == 3*70-2: #3*72:
            #     mask = (distance_list_raw[t-1]>=5)&(distance_list_raw[t-1]<=10)
                
            #     data_masked = pressure_list_raw[t-1][mask]
            #     CoV = np.std(data_masked)/np.nanmean(data_masked)
            #     print (len(data_masked))
            #     print (len(distance_list_raw[t-1]))
            #     print (CoV)
                
            #     plt.text(0.05,0.93,s=f"CoV (72h; 5-10r): {np.around(CoV,3)}",
            #              transform=plt.gca().transAxes, zorder=200,
            #              fontsize=11,c="darkred")
            
        
        # else same color for all here
        else: 
           for ci,t in enumerate(timesteps_scatter):
            if t == None: #do not plot None elements
                  continue
            plt.scatter(distance_list_raw[t-1],displacement_list_raw[t-1],s=marker_size_scatter,zorder=20,c = color_raw) # deformation and image data have a length difference of 1
        
        
    # PLOT THE MEAN RAW DATA
    if color_list and label_list:
        for ci,t in enumerate(timesteps):
            if plot_means == False:
                continue
            # calculate the mean in distance windows for timesteps
            mean_distance = []
            mean_displacement = []
            for i in range(distance[0],int(np.max(distance_list_raw[t-1]))):
                mean_distance.append(i+0.5)
                mean_disp = np.nanmean(displacement_list_raw[t-1][(distance_list_raw[t-1]>=i) & (distance_list_raw[t-1]<i+1)])
                mean_displacement.append(mean_disp)
            plt.plot(mean_distance,mean_displacement,"o-",ms=marker_size_mean, zorder=2000, 
                     linewidth=linewidth,  markerfacecolor="w",  markeredgecolor=color_list[ci], 
                     markeredgewidth=2 , label=label_list[ci], color="w" )     
        ax = plt.legend( markerscale=0.4, fontsize=6.5,loc="upper right")
        ax.set_zorder(2000)
    else: 
         for ci,t in enumerate(timesteps):
            if plot_means == False:
              continue
            # calculate the mean in distance windows for timesteps
            mean_distance = []
            mean_displacement = []
            for i in range(distance[0],int(np.max(distance_list_raw[t-1]))):
                mean_distance.append(i+0.5)
                mean_disp = np.nanmean(displacement_list_raw[t-1][(distance_list_raw[t-1]>=i) & (distance_list_raw[t-1]<i+1)])
                mean_displacement.append(mean_disp)
            plt.plot(mean_distance,mean_displacement,"o-",ms=marker_size_mean,zorder=2000,
                     linewidth=linewidth, markerfacecolor="w",color ="w",
                     markeredgecolor=color_raw, markeredgewidth=2)   
        #plt.xlim(2,20);plt.ylim(1e-4,3) 

    return 



import numpy as np
import pandas as pd
from tqdm import tqdm
from glob import glob
from natsort import natsorted
import matplotlib.cm 
import matplotlib.pyplot as plt
import os
from .simulation import load_lookup_functions
from .utils import load





def reconstruct(folder, lookupfile, muperpixel, outfile=None, r_min=2, angle_filter=20, r_max=None,  continuous_radii = False):
    """
    - folder: contains PIV results
    - lookupfile: Material look-up table for force reconstruction
    - muperpixel: Pixelsize
    - outfile: can be used to change the standard name of the excel file 
    - r_min: Specifies the minimal distance in which deformations are evaluated. 
    By default 2 radii to avoid close-range effects.
    - angle_filter:  If 'None' all Deformations are taken into account, else 
    deformations where the angle to the spheroid axis is smaller than the specified
    angle (in degree) are taken into account for a more robust evaluation. 
    Default is 20 degree.
    - continious radii: If set true the spheroid radii is updated at each timepoint for
    the force reconstruction. Default uses the intial radii for all time points to 
    reduce errors due to fluctuation of segmentation
    """
    # get filepaths for PIV results
    # load the segmentations
    seg_files = natsorted(glob(folder+'/seg*.npy'))
    
    # load deformations
    # look for accumulated deformation files (new standard)
    d_accumulated_files = natsorted(glob(folder+'/def*.npy'))
    # look also for not-accummulated deformations (old standard)
    d_notaccumulated_files = natsorted(glob(folder+'/dis*.npy'))   
    # if not-accumulated deformations are found chose different mode
    if len(d_notaccumulated_files) > len(d_accumulated_files):
        accumulated = False
        dis_files = d_notaccumulated_files
        print("Found not-accumulated deformation files (old standard) and will conduct calculations accordingly.")
    # else do the calcualtion with accumulated d eformations already
    else:
        accumulated = True
        dis_files = d_accumulated_files
       
    
    # load lookup table
    get_displacement, get_pressure = load_lookup_functions(lookupfile)

    # initial spheroid radius and surface (used for force estimation)
    r0 = load(seg_files[0])['radius']
    A0 = 4 * np.pi * (r0 * (10 ** -6)) ** 2

    # initialize result dictionary
    results = {'pressure_mean': [], 'pressure_median': [], 'pressure_std': [],
               'contractility_mean': [], 'contractility_median': [], 'contractility_std': [],
               'pressure_min': [], 'pressure_max': [], 'contractility_min': [], 'contractility_max': [],
               'angle_min': [], 'angle_max': []}

    # create dict with all angles
    angles_dict = {str(a): [] for a in range(-180, 180, 5)}
    
    u_sum = None
    v_sum = None

    # loop over series of PIV results
    for (dis_file, seg_file) in tqdm(zip(dis_files, seg_files)):
        dis = load(dis_file)
        seg = load(seg_file)

        x_rav = np.ravel(dis['x'])
        y_rav = np.ravel(dis['y'])

        #  get deformations - sum up if we have not-accummulated deformations (old standard)
        if accumulated == False:
            try:
                u_sum += np.ravel(dis['u'])
                v_sum += np.ravel(dis['v'])
            except:
                u_sum = np.ravel(dis['u'])
                v_sum = np.ravel(dis['v'])
        # else read in accummulated deformations directly (new standard)
        else:
            u_sum = np.ravel(dis['u'])
            v_sum = np.ravel(dis['v'])
                
                
        cx, cy = seg['centroid']
        
        # update radius if contious radii option is active
        if continuous_radii == True:
            r0 = seg['radius']
        # else use always the intiial timepoint          
        distance, displacement, angle, pressure = infer_pressure(x_rav, y_rav, u_sum, v_sum, cx, cy, r0, get_pressure , angle_filter=angle_filter)
        mask = distance > r_min
        if r_max:
            mask = (r_max > distance) & (distance > r_min)

        pr_angle = []
        pr_median = []
        for alpha in range(-180, 180, 5):
            mask2 = (angle >= (alpha-5)*np.pi/180.) & (angle < (alpha+5)*np.pi/180.)
            pr_angle.append(alpha)
            pr_median.append(np.nanmedian(pressure[mask & mask2]))
        
        # assign pressure values
        pressure_mean = np.nanmean(pressure[mask])
        pressure_median = np.nanmedian(pressure[mask])
        pressure_std = np.nanstd(pressure[mask], ddof=1)    
        # search maximal / minimal value
        try:
            i_min = np.nanargmin(pr_median)
            alpha_min = pr_angle[i_min]
            pressure_min = pr_median[i_min]
    
            i_max = np.nanargmax(pr_median)
            alpha_max = pr_angle[i_max]
            pressure_max = pr_median[i_max]
        # assign nan values if not possible (for example if only same values esixt at all locations due to negative/non-simulated strain        
        except: 
            i_min,alpha_min,pressure_min = np.nan,np.nan,np.nan
            i_max,alpha_max,pressure_max = np.nan,np.nan,np.nan


        contractility_mean = pressure_mean*A0*(muperpixel**2.)*(10**6)  # unit: µN
        contractility_median = pressure_median*A0*(muperpixel**2.)*(10**6)  # unit: µN
        contractility_std = pressure_std*A0*(muperpixel**2.)*(10**6)  # unit: µN

        contractility_min = pressure_min*A0*(muperpixel**2.)*(10**6)  # unit: µN
        contractility_max = pressure_max*A0*(muperpixel**2.)*(10**6)  # unit: µN

        results['pressure_mean'].append(pressure_mean)
        results['pressure_median'].append(pressure_median)
        results['pressure_std'].append(pressure_std)

        results['contractility_mean'].append(contractility_mean)
        results['contractility_median'].append(contractility_median)
        results['contractility_std'].append(contractility_std)

        results['pressure_min'].append(pressure_min)
        results['pressure_max'].append(pressure_max)

        results['contractility_min'].append(contractility_min)
        results['contractility_max'].append(contractility_max)

        results['angle_min'].append(alpha_min)
        results['angle_max'].append(alpha_max)
        
        # append pressures for all angle data
        for i,a in enumerate(angles_dict):
            angles_dict[a].append(pr_median[i])

    df = pd.DataFrame.from_dict(results)
    df.columns = ['Mean Pressure (Pa)',
                  'Median Pressure (Pa)',
                  'St.dev. Pressure (Pa)',
                  'Mean Contractility (µN)',
                  'Median Contractility (µN)',
                  'St.dev. Contractility (µN)',
                  'Minimal Median Pressure (Pa)',
                  'Maximal Median Pressure (Pa)',
                  'Minimal Median Contractility (µN)',
                  'Maximal Median Contractility (µN)',
                  'Angle of minimal Pr./Contr. (deg)',
                  'Angle of maximal Pr./Contr. (deg)']

    if outfile is not None:
        df.to_excel(outfile)
    else:
        df.to_excel(folder+'//result.xlsx')
     
    # save pressures for all angles
    an = pd.DataFrame.from_dict(angles_dict)
    an.columns = [a for a in range(-180, 180, 5)]
    if outfile is not None:
        an.to_excel(outfile[:-5]+'_angles.xlsx')
    else:
        an.to_excel(folder+'//result_angles.xlsx')    
          
    # save a copy of the lookup table by default
    from shutil import copyfile
    copyname = "AppliedLookupTable.pkl"
    copyfile(lookupfile, os.path.join(folder,copyname))
    
    # save parameter file 
    import yaml
    dict_file = {'Force' :   {'folder': [folder], 'lookupfile': [lookupfile], 'muperpixel': [muperpixel], 
              'r_min': [str(r_min)], 'r_max': [str(r_max)],
              'angle_filter': [str(angle_filter)], 
              'continuous_radii': [continuous_radii], 'accumulated': [str(accumulated)],
              'Copy of applied Lookuptable': [copyname]},     
          }  
    with open(os.path.join(folder, 'parameters_force.yml'), 'w') as yaml_file:
        yaml.dump(dict_file, yaml_file, default_flow_style=False)
  
    
        
    return df    
 



def infer_pressure(x_rav, y_rav, u_rav, v_rav, x_sph, y_sph, r_sph, get_pressure, angle_filter = 20):
    """
    - angle_filter:  If 'None' all Deformations are taken into account, else 
    deformations where the angle to the spheroid axis is smaller than the specified
    angle (in degree) are taken into account for a more robust evaluation. 
    Default is 20 degree.
    """
    # compute relative distances and angle
    dx = (x_rav[~np.isnan(u_rav)] - x_sph) / r_sph
    dy = (y_rav[~np.isnan(u_rav)] - y_sph) / r_sph
    distance = np.sqrt(dx ** 2 + dy ** 2)
    angle = np.arctan2(dy, dx)

    # compute relative deformation
    u_rav2 = u_rav[~np.isnan(u_rav)] / r_sph
    v_rav2 = v_rav[~np.isnan(u_rav)] / r_sph
    
    # project deformation to spheroid axis
    m = (np.array([dx, dy]).T) / np.expand_dims(distance, axis=1)
    d = np.array([u_rav2, v_rav2]).T
    displacement = np.array([-np.dot(di, mi) for di, mi in zip(d, m)])
    
    # optional filter deformations that are not alliged with axis 
    if angle_filter is not None:
        abs = np.sqrt(u_rav2**2. + v_rav2**2.)
        mask = displacement/abs > np.cos(2*np.pi * angle_filter/360)
        distance = distance[mask]
        angle = angle[mask]
        displacement = displacement[mask]
        
    # determine pressure with deformation distance look-up tables
    pressure = get_pressure(distance, displacement)

    return [distance, displacement, angle, pressure]


# Evaluate Angle dependet Pressures
def angle_analysis(folder, output, n_max=None,save_plot=True, small_pressure = False, fontsize=7, name_of_resultfile='result_angles.xlsx', dt=None, angle_legend=False, 
                   path_of_resultfile=None):
    """
    Evaluate angles over time for an spheroid (needs folder where 'reconstruct' function was used) and stores 
    the results in the output folder.
    - N_max may define the last image to evaluate 
    - Small_pressure true returns plots that are scaled to 100 Pa whereas false returns plots that are scaled up to 1000 Pa
    - Evaluated excel sheet with angle data is automatically searched in folder, if name differs from  'result_angles.xlsx' it can be specified using name_of_resultfile
    - dt is time between consecutive images in seconds; If given a timestamp will be displayed 
    - angle_legend will draw a legend with angles if active 
    """
    from datetime import timedelta
    
    # read in angle pressures
    if path_of_resultfile:
        angles = pd.read_excel(path_of_resultfile)
    else:
        angles  = pd.read_excel(folder + '//'+ name_of_resultfile)
        
    # read in plots
    plots = glob(folder+'//plot*.png')
    # make folder if not existing
    if not os.path.exists(output):
        os.makedirs(output)
    # list of evaluated angles
    angle_list = [a for a in range(-180, 180, 5)]
    # make figure
    f, (ax1, ax2) = plt.subplots(1, 2, figsize=(9,5))
            
    #second plot with only the left plot
    fb,axb = plt.subplots( figsize=(3,3))  
    # for color norm
    max_pr = np.nanmax([np.nanmax(np.array(angles[z])[:n_max]) for z in angle_list])  
    plt.style.use('dark_background')
    # make result dictionary
    res_angles = {'mean_pr_angles (Pa)': [], 'sd_pr_angles (Pa)': [], 'CoV': []}
    
    # loop over all timesteps
    for t in tqdm(range(len(angles[0][:n_max]))):                   
        pressures = []
        # go through different angles 
        for z in angle_list:        
             # append for each angle the pressure of the current timestep
             pressures.append(np.array(angles[z])[t])    
                
        # norm colors for each image individually    viridis/RdBu      
        colors = matplotlib.cm.get_cmap('viridis')(pressures / max_pr ) 
        colorcycle = 'darkred'
    
        # do some statistics
        mean = np.round(np.nanmean(pressures))
        sd = np.round(np.nanstd(pressures))
        CoV = np.round(np.nanstd(pressures)/np.nanmean(pressures), 2)
        res_angles['mean_pr_angles (Pa)'].append(mean)
        res_angles['sd_pr_angles (Pa)'].append(sd)
        res_angles['CoV'].append(CoV)
        
  
        if save_plot:
            if angle_legend == True:
				# position of label text
                shift = 0.9
                posx_legend = [np.cos(i*np.pi/180) * shift for i in angle_list] 
                posy_legend = [np.sin(i*np.pi/180) * shift for i in angle_list] 
				# plot text
                [ax1.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=4, horizontalalignment='center') for i in range(len(angle_list))] 
                [axb.text(posx_legend[i],posy_legend[i], str(angle_list[i]), color='darkred', fontsize=3.6, horizontalalignment='center') for i in range(len(angle_list))]   
						
 			#ax1.scatter(posx_legend,posy_legend, s=22, c=colors)
 			# Mode that shows and norms up to 1000 Pa
            if small_pressure == False:
			
				# now use 1000 as maximum
                x = [np.cos(i*np.pi/180) for i in angle_list] * (pressures / np.array([1000]*len(pressures)) ) 
                y = [np.sin(i*np.pi/180) for i in angle_list] * (pressures / np.array([1000] *len(pressures)) ) 
				# combined figure
				# plot circles
                circle_100pa = plt.Circle((0, 0), 100/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_250pa = plt.Circle((0, 0), 250/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_500pa = plt.Circle((0, 0), 500/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_750pa = plt.Circle((0, 0), 750/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                ax1.text(0.5, 0.57, '100 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
                ax1.text(0.5, 0.65, '250 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
                ax1.text(0.5, 0.77, '500 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
                ax1.text(0.5, 0.9, '750 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
                ax1.add_artist(circle_100pa)
                ax1.add_artist(circle_250pa)
                ax1.add_artist(circle_500pa)
                ax1.add_artist(circle_750pa)
                ax1.axis('equal')  
                ax1.scatter(x,y, s=22, c=colors)
                ax1.set_xlim([-1, 1])
                ax1.set_ylim([-1, 1])
                ax1.axis('off')
				# annotate CoV    
                ax1.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV), 
				 horizontalalignment='center',
				 verticalalignment='center',
				 transform = ax1.transAxes, fontsize=9.2)  
				
				# annotate time if timestep is given
                if dt is not None:
                    ax1.text(0.03, 0.98,str(timedelta(seconds=t*dt)), 
                             horizontalalignment='center',
                             verticalalignment='center',
                             transform = ax1.transAxes, fontsize=12)  

					
				# show quiver plot on the right side
                plot = plt.imread(plots[t])
                ax2.imshow(plot)
                plt.axis('off')
                ax2.axis('off')
				# save figure
                f.savefig(output+'//plot_{}.png'.format(str(t).zfill(4)), dpi=200)
                ax1.cla()
                ax2.cla()
				# Single figure
				# plot circles
                circle_100pa = plt.Circle((0, 0), 100/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_250pa = plt.Circle((0, 0), 250/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_500pa = plt.Circle((0, 0), 500/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_750pa = plt.Circle((0, 0), 750/ 1000 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                axb.text(0.5, 0.57, '100 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
                axb.text(0.5, 0.65, '250 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
                axb.text(0.5, 0.77, '500 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
                axb.text(0.5, 0.9, '750 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
                axb.add_artist(circle_100pa)
                axb.add_artist(circle_250pa)
                axb.add_artist(circle_500pa)
                axb.add_artist(circle_750pa)   
                axb.axis('equal')  
                axb.scatter(x,y, s=22, c=colors)
                axb.set_xlim([-1, 1])
                axb.set_ylim([-1, 1])
                axb.axis('off')
                plt.savefig(output+'//single_{}.png'.format(str(t).zfill(4)), dpi=200)
                axb.cla()
                
                # that shows and norms up to 100 Pa
            if small_pressure == True:
				# now use 100 as maximum
                x = [np.cos(i*np.pi/180) for i in angle_list] * (pressures / np.array([100]*len(pressures)) ) 
                y = [np.sin(i*np.pi/180) for i in angle_list] * (pressures / np.array([100] *len(pressures)) ) 
				# combined figure
				# plot circles
                circle_a = plt.Circle((0, 0), 10/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_b = plt.Circle((0, 0), 25/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_c = plt.Circle((0, 0), 50/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                ax1.text(0.5, 0.57, '10 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
                ax1.text(0.5, 0.65, '25 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
                ax1.text(0.5, 0.77, '50 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=ax1.transAxes,fontsize=fontsize)
                ax1.add_artist(circle_a)
                ax1.add_artist(circle_b)
                ax1.add_artist(circle_c)
                ax1.axis('equal')  
                ax1.scatter(x,y, s=22, c=colors)
                ax1.set_xlim([-1, 1])
                ax1.set_ylim([-1, 1])
                ax1.axis('off')
				# annotate CoV    
                ax1.text(0.01, 0.91,'Coefficient of Variation: '+str(CoV), 
                         horizontalalignment='center',
                             verticalalignment='center',
                            transform = ax1.transAxes, fontsize=9.2)  
				
				# annotate time if timestep is given
                if dt is not None:
                    ax1.text(0.03, 0.98,str(timedelta(seconds=t*dt)), 
                             horizontalalignment='center',
                             verticalalignment='center',
                             transform = ax1.transAxes, fontsize=12)  
				# show quiver plot on the right side
                plot = plt.imread(plots[t])
                ax2.imshow(plot)
                plt.axis('off')
                ax2.axis('off')
				# save figure
                f.savefig(output+'//plot_{}.png'.format(str(t).zfill(4)), dpi=200)
                ax1.cla()
                ax2.cla()
				
				# single figure
				# plot circles       
                circle_a = plt.Circle((0, 0), 10/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_b = plt.Circle((0, 0), 25/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                circle_c = plt.Circle((0, 0), 50/ 100 , color=colorcycle, zorder=10, fill=False, linestyle='--')
				#circle_750pa = plt.Circle((0, 0), 100/ 300 , color=colorcycle, zorder=10, fill=False, linestyle='--')
                axb.text(0.5, 0.57, '10 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
                axb.text(0.5, 0.65, '25 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
                axb.text(0.5, 0.77, '50 Pa', color = colorcycle ,horizontalalignment='center', verticalalignment='center', transform=axb.transAxes,fontsize=fontsize)
                axb.add_artist(circle_a)
                axb.add_artist(circle_b)
                axb.add_artist(circle_c)
                axb.axis('equal')  
                axb.scatter(x,y, s=22, c=colors)
                axb.set_xlim([-1, 1])
                axb.set_ylim([-1, 1])
                axb.axis('off')
                plt.savefig(output+'//single_{}.png'.format(str(t).zfill(4)), dpi=200)
                axb.cla()


    # save excel file     
    ae = pd.DataFrame.from_dict(res_angles)
    ae.columns = ['mean_pr_angles (Pa)',
                  'sd_pr_angles (Pa)',
                  'CoV',]

    ae.to_excel(output+'/angles_eval.xlsx')

    return ae








import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from glob import glob
from tqdm import tqdm
from natsort import natsorted
from scipy.ndimage import generic_filter
from skimage.morphology import disk
from scipy.interpolate import LinearNDInterpolator
from saenopy import macro
from saenopy.materials import SemiAffineFiberMaterial, LinearMaterial
from scipy import interpolate
import matplotlib 
import os


def get_displ(x_rav, y_rav, u_rav, v_rav, x_sph, y_sph):
    dx = (x_rav - x_sph)
    dy = (y_rav - y_sph)

    distance = np.sqrt(dx ** 2 + dy ** 2)
    angle = np.arctan2(dy, dx)
    m = (np.array([dx, dy]).T) / np.expand_dims(distance, axis=1)
    d = np.array([u_rav, v_rav]).T
    displacement = np.array([-np.dot(di, mi) for di, mi in zip(d, m)])

    return [distance, angle, displacement]


def create_strain_maps(folder, delta, outfolder='strain-maps', radius=2, i_max=None, vmin=1, vmax=1.5, ticks_strains=[ 1, 1.1 , 1.2 , 1.3, 1.4, 1.5]):
    """
    Creates 2D Strain maps for a time series. 
    (and single plot with mean and maximal strain for eacht time step)
    
    folder: containing the defromation series (output folder of compute_displacement_series)
    outfolder: output of Strain maps 
    delta: Windowsize (px) as it was used in the evaluation of deformations for correct strain
    radius: used to smooth the deformation field    
    i_max: can be used to evaluate only first i timesteps
    """
    # creates folder if it doesn't exist
    if not os.path.exists(outfolder):
        os.makedirs(outfolder)
    # creates folder if it doesn't exist
    if not os.path.exists(os.path.join(outfolder,"plots")):
        os.makedirs(os.path.join(outfolder,"plots"))   
        
    
    # load segmentations
    seg_files = natsorted(glob(folder+'/seg*.npy'))
    

    # load deformations
    # look for accumulated deformation files (new standard)
    d_accumulated_files = natsorted(glob(folder+'/def*.npy'))
    # look also for not-accummulated deformations (old standard)
    d_notaccumulated_files = natsorted(glob(folder+'/dis*.npy'))   
    # if not-accumulated deformations are found chose different mode
    if len(d_notaccumulated_files) > len(d_accumulated_files):
        accumulated = False
        dis_files = d_notaccumulated_files
        print("Found not-accumulated deformation files (old standard) and will conduct calculations accordingly.")
    # else do the calcualtion with accumulated d eformations already
    else:
        accumulated = True
        dis_files = d_accumulated_files
    

    dis = np.load(dis_files[0], allow_pickle="True").item()

    x_rav, y_rav = np.ravel(dis['x']), np.ravel(dis['y'])
    u_rav, v_rav = np.zeros_like(np.ravel(dis['u'])), np.zeros_like(np.ravel(dis['v']))

    if not i_max:
        i_max= len(dis_files)
    median_strain = []
    max_strain = []
    max_defo = []
    median_defo = []
    for i in tqdm(range(i_max)):
        # get displacement field
        dis = np.load(dis_files[i], allow_pickle="True").item()
        
        # get deformations
        # sum up if we have not-accummulated deformations (old standard)
        if accumulated == False:
            try:
                u_rav += np.ravel(dis['u'])
                v_rav += np.ravel(dis['v'])
            except:
                u_rav = np.ravel(dis['u'])
                v_rav = np.ravel(dis['v'])
        # else read in accummulated deformations directly (new standard)
        else:
            u_rav = np.ravel(dis['u'])
            v_rav = np.ravel(dis['v'])
        

        seg = np.load(seg_files[i], allow_pickle="True").item()
        x_sph, y_sph = seg['centroid']

        dist, angle, displ = get_displ(x_rav, y_rav, u_rav, v_rav, x_sph, y_sph)

        # make sure spheroid mask is visible at the end
        dist[np.isnan(u_rav)] = np.nan

        # smooth displacement field
        displ = displ.reshape(dis['x'].shape).copy()
        displ = generic_filter(displ, np.nanmean, footprint=disk(radius), mode='nearest')

        # linear interpolation of discrete displacement field
        displ = np.ravel(displ)
        x, y = np.ravel(dis['x']), np.ravel(dis['y'])

        mask = ~(np.isnan(x) | np.isnan(y) | np.isnan(displ))

        x2 = x[mask]
        y2 = y[mask]
        displ2 = displ[mask]

        f = LinearNDInterpolator(np.array([x2, y2]).T, displ2)

        # compute strain
        x_old = (dist * np.cos(angle) + x_sph).reshape(dis['x'].shape)
        y_old = (dist * np.sin(angle) + y_sph).reshape(dis['y'].shape)
        dist_old = f(x_old, y_old)

        x_new = ((dist + delta) * np.cos(angle) + x_sph).reshape(dis['x'].shape)
        y_new = ((dist + delta) * np.sin(angle) + y_sph).reshape(dis['y'].shape)
        dist_new = f(x_new, y_new)

        strain = 1 + (dist_old - dist_new) / delta

        #print (strain)
        print ("")
        print ("Max. Strain: "+str(np.nanmax(strain)))
        max_defo.append(np.nanmax(np.sqrt(u_rav**2 + v_rav**2 )))
        median_defo.append(np.nanmedian(np.sqrt(u_rav**2 + v_rav**2 )))
        max_strain.append(np.nanmax(strain))
        median_strain.append(np.nanmedian(strain))
        plt.figure(figsize=(8, 6))
        height = strain.shape[0]
        width = strain.shape[1]
        # plot strain map            
        plt.imshow(strain, vmin=vmin, vmax=vmax, cmap='magma', extent=[0, width, height, 0], origin='lower')
        cb = plt.colorbar(fraction=0.025, label='Strain', ticks=ticks_strains)
            
        #cb.ax.set_yticklabels(['$\mathregular{\leq}$ 300 Pa', '400 Pa', '500 Pa', '$\mathregular{\geq}$ 600 Pa'])
        plt.xlim((0, width))
        plt.ylim((0, height))
        plt.axis('off')
        plt.gca().get_xaxis().set_visible(False)
        plt.gca().get_yaxis().set_visible(False)
        plt.tight_layout()
        plt.savefig(outfolder+'/strain'+str(i+1).zfill(6)+'.png', dpi=250)
        plt.close()
   
    # save last strain map + deformations interpolated
    np.save(os.path.join(outfolder,"strain_end.npy"),strain)
    np.save(os.path.join(outfolder,"x_end.npy"),x2)
    np.save(os.path.join(outfolder,"y_end.npy"),y2)
    np.save(os.path.join(outfolder,"displ_end.npy"),displ2)
    # raw accumulated data
    np.save(os.path.join(outfolder,"x_rav_end.npy"),x_rav)
    np.save(os.path.join(outfolder,"y_rav_end.npy"),y_rav)
    np.save(os.path.join(outfolder,"u_rav_end.npy"),u_rav)
    np.save(os.path.join(outfolder,"v_rav_end.npy"),v_rav)  
    
    
    # plot and save overview
    np.savetxt(os.path.join(outfolder,"plots")+'/median_strain.txt', median_strain)
    np.savetxt(os.path.join(outfolder,"plots")+'/max_strain.txt', max_strain)  
    np.savetxt(os.path.join(outfolder,"plots")+'/max_defo.txt', max_defo)  
    np.savetxt(os.path.join(outfolder,"plots")+'/median_defo.txt', median_defo)  
    
    # deformations  
    plt.figure(figsize=(6,2))
    plt.grid()
    plt.xlabel("timesteps")
    plt.ylabel("deformations (px)")
    plt.plot(max_defo)
    plt.tight_layout()
    plt.savefig(os.path.join(outfolder,"plots")+'/max_defo.png', dpi=300)
    plt.close()
    # median strain    
    plt.figure(figsize=(6,2))
    plt.grid()
    plt.xlabel("timesteps", fontsize=7)
    plt.ylabel("median deformations (px)", fontsize=7)
    plt.plot(median_defo)
    plt.tight_layout()
    plt.savefig(os.path.join(outfolder,"plots")+'/median_deformation.png', dpi=300)
    plt.close()
    # median strain    
    plt.figure(figsize=(6,2))
    plt.grid()
    plt.xlabel("timesteps")
    plt.ylabel("median strain")
    plt.plot(median_strain)
    plt.tight_layout()
    plt.savefig(os.path.join(outfolder,"plots")+'/median_strain.png', dpi=300)
    plt.close()
    # max strain
    # max strain
    plt.figure(figsize=(6,2))
    plt.grid()
    plt.xlabel("timesteps")
    plt.ylabel("max strain")
    plt.plot(max_strain)
    plt.tight_layout()
    plt.savefig(os.path.join(outfolder,"plots")+'/max_strain.png', dpi=300)
    plt.close()
    return 


def create_stiffness_maps(folder, delta, outfolder='stiffness-maps', k0=1645, eps_s=0.0075, d_s=0.033, d_0=0.0008,
                           radius=2, i_max=None):
    
    """
    Creates 2D stiffness maps for a time series. 
    (and single plot with mean and maximal stiffness for eacht time step)
    
    folder: containing the defromation series (output folder of compute_displacement_series)
    outfolder: output of Strain maps 
    delta: Windowsize (px) as it was used in the evaluation of deformations for correct strain
    radius: used to smooth the deformation field    
    i_max: can be used to evaluate only first i timesteps
    k0, eps_s,d_s,_do: material parameters to compute the stiffness

    """
    # creates folder if it doesn't exist
    if not os.path.exists(outfolder):
        os.makedirs(outfolder)
    # creates folder if it doesn't exist
    if not os.path.exists(os.path.join(outfolder,"plots")):
        os.makedirs(os.path.join(outfolder,"plots"))   
        
    # fiber material
    material = SemiAffineFiberMaterial(k0, d_0, eps_s, d_s)
    # loead extensional stress data and get stiffness - might need to adjust range
    epsilon = np.arange(-1, 1.9, 0.0001)
    x, y = macro.getExtensionalRheometerStress(epsilon, material)
    get_stiffness = interpolate.interp1d(x[1:], (y[1:]-y[:-1])/(x[1:]-x[:-1]))
        
    # load in displacement data and ssheroid mask
    # load segmentations
    seg_files = natsorted(glob(folder+'/seg*.npy'))

    # load deformations
    # look for accumulated deformation files (new standard)
    d_accumulated_files = natsorted(glob(folder+'/def*.npy'))
    # look also for not-accummulated deformations (old standard)
    d_notaccumulated_files = natsorted(glob(folder+'/dis*.npy'))   
    # if not-accumulated deformations are found chose different mode
    if len(d_notaccumulated_files) > len(d_accumulated_files):
        accumulated = False
        dis_files = d_notaccumulated_files
        print("Found not-accumulated deformation files (old standard) and will conduct calculations accordingly.")
    # else do the calcualtion with accumulated d eformations already
    else:
        accumulated = True
        dis_files = d_accumulated_files
    
    

    dis = np.load(dis_files[0], allow_pickle="True").item()

    x_rav, y_rav = np.ravel(dis['x']), np.ravel(dis['y'])
    u_rav, v_rav = np.zeros_like(np.ravel(dis['u'])), np.zeros_like(np.ravel(dis['v']))
    
    # take all timesteps if not specified
    if not i_max:
        i_max= len(dis_files)
    
    # loop over all times
    median_stiffness = []
    max_stiffness = []    
    for i in tqdm(range(i_max)):
        # get displacement field
        dis = np.load(dis_files[i], allow_pickle="True").item()
      
        # get deformations
        # sum up if we have not-accummulated deformations (old standard)
        if accumulated == False:
            try:
                u_rav += np.ravel(dis['u'])
                v_rav += np.ravel(dis['v'])
            except:
                u_rav = np.ravel(dis['u'])
                v_rav = np.ravel(dis['v'])
        # else read in accummulated deformations directly (new standard)
        else:
            u_rav = np.ravel(dis['u'])
            v_rav = np.ravel(dis['v'])
   

        seg = np.load(seg_files[i], allow_pickle="True").item()
        x_sph, y_sph = seg['centroid']

        dist, angle, displ = get_displ(x_rav, y_rav, u_rav, v_rav, x_sph, y_sph)

        # make sure spheroid mask is visible at the end
        dist[np.isnan(u_rav)] = np.nan

        # smooth displacement field
        displ = displ.reshape(dis['x'].shape).copy()
        displ = generic_filter(displ, np.nanmean, footprint=disk(radius), mode='nearest')

        # linear interpolation of discrete displacement field
        displ = np.ravel(displ)
        x, y = np.ravel(dis['x']), np.ravel(dis['y'])

        mask = ~(np.isnan(x) | np.isnan(y) | np.isnan(displ))

        x2 = x[mask]
        y2 = y[mask]
        displ2 = displ[mask]

        f = LinearNDInterpolator(np.array([x2, y2]).T, displ2)

        # compute strain
        x_old = (dist * np.cos(angle) + x_sph).reshape(dis['x'].shape)
        y_old = (dist * np.sin(angle) + y_sph).reshape(dis['y'].shape)
        dist_old = f(x_old, y_old)

        x_new = ((dist + delta) * np.cos(angle) + x_sph).reshape(dis['x'].shape)
        y_new = ((dist + delta) * np.sin(angle) + y_sph).reshape(dis['y'].shape)
        dist_new = f(x_new, y_new)

        strain = 1 + (dist_old - dist_new) / delta

        # compute stiffening
        stiffness = get_stiffness(strain)
        log_stiffness = np.log10(stiffness)
        #stiffness = np.zeros_like(strain) + np.nan  # initalize with linear stiffness
        #stiffness[strain <= 0] = k0 * np.exp(strain[strain <= 0] / d_0)  # buckling
        #stiffness[(strain > 0) & (strain < eps_s)] = k0  # linear regime
        #stiffness[strain >= eps_s] = k0 * np.exp((strain[strain >= eps_s] - eps_s) / d_s)  # strain stiffening
        # plot result
        plt.figure(figsize=(8, 6))
        height = log_stiffness.shape[0]
        width = log_stiffness.shape[1]
        plt.imshow(log_stiffness, vmin=2, vmax=4, cmap='inferno', extent=[0, width, height, 0], origin='lower')
        cb = plt.colorbar(fraction=0.025, label='Matrix stiffness', ticks=np.log10([100, 250, 500, 1000, 2500, 5000, 10000]))
        cb.ax.set_yticklabels(['$\mathregular{\leq}$ 100 Pa', '250 Pa', '500 Pa', '1.0 kPa', '2.5 kPa', '5.0 kPa', '$\mathregular{\geq}$ 10.0 kPa'])
        plt.xlim((0, width))
        plt.ylim((0, height))
        plt.axis('off')
        plt.gca().get_xaxis().set_visible(False)
        plt.gca().get_yaxis().set_visible(False)
        plt.tight_layout()
        plt.savefig(outfolder+'/stiffening'+str(i+1).zfill(6)+'.png', dpi=250)
        print('')
        print("Max. Stiffness: "+str(np.nanmax(stiffness)))
        print('')
        median_stiffness.append(np.nanmedian(stiffness))
        max_stiffness.append(np.nanmax(stiffness))
        plt.close()
        
     # plot and save overview    
    np.savetxt(os.path.join(outfolder,"plots")+'/median_stiffness.txt', median_stiffness)
    np.savetxt(os.path.join(outfolder,"plots")+'/max_stiffness.txt', max_stiffness)
    # median stiffness
    plt.figure(figsize=(6,2))
    plt.grid()
    plt.xlabel("timesteps")
    plt.ylabel("median stiffness")
    plt.plot(median_stiffness)
    plt.tight_layout()
    plt.savefig(os.path.join(outfolder,"plots")+'/median_stiffness.png', dpi=300)
    plt.close()
    # max stiffness
    plt.figure(figsize=(6,2))
    plt.grid()
    plt.xlabel("timesteps")
    plt.ylabel("max. stiffness")
    plt.plot(max_stiffness)
    plt.tight_layout()
    plt.savefig(os.path.join(outfolder,"plots")+'/max_stiffness.png', dpi=300)
    plt.close()
    return 



import numpy as np
from time import sleep


def spherical_inclusion(outfile, r_inner=100, r_outer=10000, length_factor=0.05):
    """
    Creates a spherical bulk mesh with a centered spherical inclusion.
    
    Args:
        outfile(str): File path to save mesh file (add .msh ending)
        r_inner(float): Radius of the spherical inclusion (in µm)
        r_outer(float): Outer radius of the bulk mesh (in µm)
        length_factor(float): Mesh element size is determined by curvature and then multipled with this factor
    """
    # Build model (OpenCascade (for boolean operators) or built-in)
    import gmsh
    model = gmsh.model
    factory = model.occ  # occ opencascade for boolean operations,  else (factory = model.geo)
    gmsh.initialize('', False)  # do not read in config files,  else (gmsh.initialize(sys.argv))
    gmsh.option.setNumber("General.Terminal", 1)
    model.add("TwoSpheres")  # Model Name
    
    # Mesh algorithm ---------------------------------------------------------------------------------------------------
    # 2D mesh algorithm (1: MeshAdapt, 2: Automatic, 5: Delaunay, 6: Frontal, 7: BAMG, 8: DelQuad)
    gmsh.option.setNumber("Mesh.Algorithm", 2)
    
    # 3D mesh algorithm (1: Delaunay, 4: Frontal, 5: Frontal Delaunay, 6: Frontal Hex, 7: MMG3D, 9: R-tree, 10: HXT)
    gmsh.option.setNumber("Mesh.Algorithm3D", 1)
    
    # Mesh size --------------------------------------------------------------------------------------------------------
    # Factor applied to all mesh element sizes, Default value: 1  
    gmsh.option.setNumber('Mesh.CharacteristicLengthFactor', length_factor)   
    
    # Element Size by Curvature (0=off 1= on, Include curvature into element mesh size)
    gmsh.option.setNumber("Mesh.CharacteristicLengthFromCurvature", 1)
    
    # Extend computation of mesh element sizes from the boundaries into the interior 
    # (for 3D Delaunay, use 1: longest or 2: shortest surface edge length; 1 by default)
    gmsh.option.setNumber("Mesh.CharacteristicLengthExtendFromBoundary", 1)
    
    # Turn off 0 for dimensionless mesh, or if 1 set mesh size (so far only possible for points..)
    gmsh.option.setNumber("Mesh.CharacteristicLengthFromPoints", 0)
    
    # Optimize the Mesh-------------------------------------------------------------------------------------------------
    # Optimize the mesh to improve the quality of tetrahedral elements, Default value: 1
    gmsh.option.setNumber("Mesh.Optimize", 1)

    # Optimize tetrahedra that have a quality below ...,  Default value: 0.3       
    gmsh.option.setNumber("Mesh.OptimizeThreshold", 0.3)
    
    # Optimize the mesh using Netgen to improve the quality of tetrahedral elements, Default value: 0
    gmsh.option.setNumber("Mesh.OptimizeNetgen", 1)

    # Mesh Output Format -----------------------------------------------------------------------------------------------
    # default=10  (1: msh, 2: unv, 10: auto, 16: vtk, 19: vrml, 21: mail, 26: pos stat, 27: stl, 28: p3d, 30: mesh, 
    #              31: bdf, 32: cgns, 33: med, 34: diff, 38: ir3, 39: inp, 40: ply2, 41: celum, 42: su2, 47: tochnog, 
    #              49: neu, 50: matlab)
    gmsh.option.setNumber("Mesh.Format", 1)
    
    # Currently select version 2 (2.2) to apply appropriate boundary conditions
    gmsh.option.setNumber('Mesh.MshFileVersion', 2.2)
    
    # Create Geometry---------------------------------------------------------------------------------------------------
    factory.addSphere(0, 0, 0, r_inner*1e-6, tag=1, angle1=-np.pi/2, angle2=np.pi/2, angle3=2*np.pi)  # inclusion
    factory.addSphere(0, 0, 0, r_outer*1e-6, tag=2, angle1=-np.pi/2, angle2=np.pi/2, angle3=2*np.pi)  # bulk
    
    # Boolean cut outer Sphere
    # cut: boolean difference   
    # Input: objectDimTags, toolDimTags, tag, removeObject, removeTool      
    # Output: outDimTags, outDimTagsMap 
    factory.cut([(3, 2)], [(3, 1)], tag=3, removeObject=True, removeTool=True)  # boolean cut
    factory.synchronize()
    
    # Define Physical Group (to apply boundary conditions later-on)
    model.addPhysicalGroup(3,[3], 4)  # (dim, tags, newtag)
    
    #Synchronize and generate mesh
    factory.synchronize()
    model.mesh.generate(3)
    
    # Save the mesh model
    gmsh.write(outfile)
    sleep(1)  # wait for file to be written to disk

    # add comments to the file so that jointforces can identify it later on
    comment_string = '$Jointforces\ninfo=This mesh was created with JOINTFORCES.\ntype=spherical_inclusion\nr_inner={}\nr_outer={}\nlength_factor={}\n$EndJointforces\n'.format(r_inner, r_outer, length_factor)

    with open(outfile, 'a') as f:
        f.write(comment_string)

    print('Saved spherical inclusion mesh at: {}'.format(outfile))


def show_mesh(mesh_file):
    """
    Opens and shows mesh file in Gmsh
    
    Args:
        mesh_file(str): File path to load mesh file
    """
    import gmsh
    gmsh.initialize('', False)  # do not read in config files
    gmsh.open(mesh_file)
    gmsh.fltk.run()



import warnings
import os
import gc
import numpy as np
import scipy.ndimage.morphology as scipy_morph
import scipy.ndimage.measurements as scipy_meas
from skimage.filters import gaussian, threshold_otsu, threshold_yen
from skimage.morphology import remove_small_objects
from skimage.exposure import adjust_gamma   
from skimage import io
from glob import glob
from tqdm import tqdm
from natsort import natsorted
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import openpiv.pyprocess
#import openpiv.process
import openpiv.filters
from roipoly import RoiPoly
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import matplotlib as mpl
import pandas as pd
from skimage.registration import phase_cross_correlation
from scipy.ndimage import shift

def enhance_contrast(img, gauss=False, gamma=None):
    img = img.astype(float)
    # apply gaussian filter 
    if gauss==True:
        img = gaussian(img, 2)
    # contrast enhancement
    img -= np.percentile(img, 0.5)
    img /= np.percentile(img, 99.5)
    img[img < 0.] = 0.
    img[img > 1.] = 1.
    img /= 256
    # gamma correction 
    if gamma is not None:
        img = adjust_gamma(img, gamma)  
    return img
        


def segment_spheroid(img, enhance=True, thres = 0.9, thres_yen= False):
    """
    Image segmentation function to create spheroid mask, radius, and position of a spheroid in a grayscale image.

    Args:
        img(array): Grayscale image as a Numpy array
        thres(float): To adjust the segmentation

    Returns:
        dict: Dictionary with keys: mask, radius, centroid (x/y)
    """
    height = img.shape[0]
    width = img.shape[1]
    
    # contrast enhancement
    if enhance:
        img = enhance_contrast(img, gauss=True)
        
    
    # flip y (to match x/y coordinates) and binarize
    if thres_yen:
        mask = img[::-1] < threshold_otsu(img) * thres  
    else:
        mask = img[::-1] < threshold_yen(img) * thres    
    
    # remove other objects
    mask = scipy_morph.binary_dilation(mask, iterations=3)
    mask = scipy_morph.binary_fill_holes(mask)
    mask = remove_small_objects(mask, min_size=1000)
    mask = scipy_morph.binary_closing(mask, iterations=3)
    mask = scipy_morph.binary_fill_holes(mask)

    # identify spheroid as the most centered object
    labeled_mask, max_lbl = scipy_meas.label(mask)
    center_of_mass = np.array(scipy_meas.center_of_mass(mask, labeled_mask, range(1, max_lbl + 1)))
    
    try:
        distance_to_center = np.sqrt(np.sum((center_of_mass - np.array([height / 2, width / 2])) ** 2, axis=1))

        mask = (labeled_mask == distance_to_center.argmin() + 1)
        
        # determine radius of spheroid
        radius = np.sqrt(np.sum(mask) / np.pi)
        
        # determine center position of spheroid
        cy, cx = center_of_mass[distance_to_center.argmin()]
    except:
        print ("Error: Could not detect spheroid. You might adjust the segmentation." )
        import sys
        sys.exit(1)
        
    # return dictionary containing spheroid information
    return {'mask': mask, 'radius': radius, 'centroid': (cx, cy)}



def custom_mask(img):
    """
    Image segmentation function to create a custom polygon mask, and evalute radius and position of the masked object.
    Need to use %matplotlib qt in jupyter notebook
    Args:
        img(array): Grayscale image as a Numpy array
    Returns:
        dict: Dictionary with keys: mask, radius, centroid (x/y)
    """
    height = img.shape[0]
    width  = img.shape[1]
    # click polygon mask interactive
    plt.ion()
    plt.imshow(img, extent=[0, width, height, 0])
    plt.text(0.5, 1.05,'Click Polygon Mask with left click, finish with right click',  fontsize=12,
         horizontalalignment='center',
         verticalalignment='center',c='darkred', transform= plt.gca().transAxes)#     transform = ax.transAxes)
    my_roi = RoiPoly(color='r')
    # Extract mask and segementation details
    mask = np.flipud(my_roi.get_mask(img))   # flip mask due to imshow 
    # determine radius of spheroid
    radius = np.sqrt(np.sum(mask) / np.pi)
    # determine center of mass
    cy, cx = scipy_meas.center_of_mass(mask)
    # hide pop-up windows   
    plt.ioff()
    # return dictionary containing spheroid information
    return {'mask': mask, 'radius': radius, 'centroid': (cx, cy)} 


def compute_displacements(window_size, img0, img1, mask1=None, cutoff=None, drift_correction=True, drift_correction_raw_img=False):
    # get image size
    height = img1.shape[0]
    width = img1.shape[1]
    
    # drift correct raw images if activated
    if drift_correction_raw_img:
     shiftvalue = phase_cross_correlation(img0,img1) 
     #np.array(Image.open(series[n-1])), np.array(Image.open(series[n]))) #, upsample_factor=10)
     img1 =  shift(img1, shiftvalue[0], mode='nearest')
     #saveim = Image.fromarray(new_im)



    # ignore ubiquitous warning messages from OpenPIV
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        # compute displacements
        ut, vt, sig2noise = openpiv.pyprocess.extended_search_area_piv(img0,   
                                                  img1,
                                                  window_size=window_size,
                                                  search_area_size = window_size,
                                                  overlap=window_size // 2,
                                                  dt=1,
                                                  sig2noise_method='peak2peak')

        # ToDo  Signal to noise filter option by threshold
        # u, v, mask = validation.sig2noise_val( ut, vt, sig2noise, threshold = 2.5 )
        # u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)

        # replace outliers
        ut, vt = openpiv.filters.replace_outliers(ut, vt, method='localmean', max_iter=3, kernel_size=1)

        # get coordinates corresponding to displacement
        x, y = openpiv.pyprocess.get_coordinates(image_size=img1.shape,
                                                 search_area_size=window_size,
                                                 overlap=window_size // 2)
    # change OpenPiv conversion for matplotlib / python arrays
    y = y[::-1]    # flip y-axis for correct orientation with image
    vt = -vt       # turn y component of deformations for correct orientation 
    
    
    # if mask is specified, replace displacements within mask with NaN
    # if cutoff is specified, replace displacements further out from the center than cutoff value (px) with NaN
    for i, xi in enumerate(x[0, :]):
        for j, yj in enumerate(y[:, 0]):
            if mask1 is not None:
                if np.any(mask1[int(yj - 0.5 * window_size):int(yj + 0.5 * window_size),
                          int(xi - 0.5 * window_size):int(xi + 0.5 * window_size)]):
                    ut[j, i] = np.nan
                    vt[j, i] = np.nan
            if cutoff is not None:   
                if np.sqrt(((xi - 0.5 * width) ** 2. + (yj - 0.5 * height) ** 2.)) > cutoff:
                    ut[j, i] = np.nan
                    vt[j, i] = np.nan

    # subtracting mean displacements acts like a drift correction
    if drift_correction:
        ut -= np.nanmean(ut)
        vt -= np.nanmean(vt)
    return {'x': x, 'y': y , 'u': ut, 'v': vt}


def displacement_plot(img, segmentation, displacements, quiver_scale=1, color_norm=75., cmap=cm.jet, s=50,
                      cbar_um_scale = None, **kwargs):
    # get image size
    height = img.shape[0]
    width = img.shape[1]
    
    x, y, u, v = displacements['x'], displacements['y'], displacements['u'], displacements['v']
    mask = segmentation['mask']
    cx, cy = segmentation['centroid']

    plt.imshow(enhance_contrast(img), cmap='Greys_r', extent=[0, width, height, 0], origin='lower')
   
    if cmap is None:
        p = plt.quiver(x, y, u, v,
                       alpha=0.8,
                       scale=quiver_scale,
                       units='xy',
                       pivot='mid',
                       **kwargs)
    # if colorbar is plotted convert deformation to um and use the given 
    # color norm as maximal deformation
    elif cbar_um_scale is not None:  
        d = (u ** 2 + v ** 2) ** 0.5
        p = plt.quiver(x, y, u, v, d*cbar_um_scale,     
                       cmap=cmap,
                       clim=[0, color_norm],
                       alpha=0.8,
                       scale=quiver_scale, #headlength =2,
                       units='xy',
                       pivot='mid', #width=7,
                       **kwargs)      
    else:
        d = (u ** 2 + v ** 2) ** 0.5
        p = plt.quiver(x, y, u, v, d,
                       clim=[0, color_norm],
                       cmap=cmap,
                       alpha=0.8,
                       scale=quiver_scale,
                       units='xy',
                       pivot='mid',
                       **kwargs)
        
    overlay = mask.astype(int) - scipy_morph.binary_erosion(mask, iterations=4).astype(int)
    overlay = np.array([overlay.T,
                        np.zeros_like(overlay).T,
                        np.zeros_like(overlay).T,
                        overlay.T]).T
 
    plt.imshow(overlay.astype(float), extent=[0, width, height, 0], zorder=1000)
    plt.scatter([cx], [cy], lw=1, edgecolors='k', c='r', s=s)
    
    plt.xlim((0, width))
    plt.ylim((0, height))

    plt.axis('off')
    plt.gca()
    plt.gca().get_xaxis().set_visible(False)
    plt.gca().get_yaxis().set_visible(False)
    
    return p


def save_displacement_plot(filename, img, segmentation, displacements, quiver_scale=1, color_norm=75., cmap=cm.jet,
                           s=50,cbar_um_scale = None,dpi=150,t=None, colorlegend="w", **kwargs):
    height = img.shape[0]
    width = img.shape[1]
    fig = plt.figure(figsize=(10 * (width / height), 10))

    displacement_plot(img, segmentation, displacements, quiver_scale, color_norm, 
                      cmap,cbar_um_scale = cbar_um_scale, **kwargs)
    ax = plt.gca()
    # make colorbar (+scalebar) if active
    if cbar_um_scale is not None: 
        f=17.5 # fontsize
        # add scalebar  
        from matplotlib_scalebar.scalebar import ScaleBar
        scalebar = ScaleBar(cbar_um_scale, "µm",length_fraction=0.1, location="lower right", 
                            box_alpha=0 ,  color=colorlegend, font_properties={'size':f}) #fixed_value=200,
        # scalebar = ScaleBar(cbar_um_scale, "µm", location="lower right", 
        #                     box_alpha=0 ,  fixed_value=500,color="k",
        #                     scale_loc="none",label_loc="none",font_properties={'size':f}) 
        plt.gca().add_artist(scalebar)
        # add timestamp if time is specified  as t
        if t is not None:
            #from datetime import timedelta

            plt.text(0.08, 0.91,  str(int(t/60)).zfill(2)+":"+str(t%60).zfill(2), 
                     color=colorlegend, fontsize=f+8, horizontalalignment='center',
                     verticalalignment='center', transform = ax.transAxes)
        # create a mappable for the colormap with same range
        norm = mpl.colors.Normalize(vmin=0,vmax=color_norm)
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        # draw the colorbar into plot
        cbaxes = inset_axes(ax, width="20%", height="2%", loc=1, borderpad=3.1) 
        #cbaxes = inset_axes(ax, width="20%", height="2%", loc=1, borderpad=-6.1)      
        cbar = plt.colorbar(sm,cax=cbaxes, orientation='horizontal')
        cbar.ax.tick_params(labelsize=f, colors =colorlegend)
        cbar.set_label(label='Deformation (µm)',fontsize=f, color=colorlegend)  
        # ticks
        cbar.set_ticks([0.,color_norm/2,color_norm])
        cbar.set_ticklabels(["0",f"{color_norm/2}",f">{color_norm}"])

        
        
    plt.savefig(filename, bbox_inches='tight', pad_inches=0, dpi=dpi)
    plt.clf()
    plt.close(fig)





def compute_displacement_series(folder, filter, outfolder, n_max=None, n_min=None,
                                image_list=None, enhance=True, window_size=70, cutoff=None, drift_correction=True,
                                drift_correction_raw_img =False,
                                plot=True, continous_segmentation = False, quiver_scale=1, color_norm=75., 
                                draw_mask = False, gamma=None, gauss=False, load_mask=None, thres_segmentation = 0.9,
                                cut_img = False, cut_img_val = (None,None,None,None), cbar_um_scale = None, dpi=150,
                                dt_min=None, cmap="turbo", colorlegend="w", callback=None, thres_yen= False):
    # see if turbo is installed
    if cmap == "turbo":
        try:
            plt.get_cmap("turbo")
        except ValueError:
            cmap = "jet"
    
    ## read in images
    ### use glob string of folder & filter by default
    if image_list is None:
        img_files = natsorted(glob(folder+'/'+filter))
    # alternativly use defined input list if specified 
    # (folder and filter might be set to None here and are ignored)
    else:
        img_files = image_list
            
    # mainimal and maximal timesteps for analysis
    if n_max is not None:
        img_files = img_files[:n_max+1]
        
    if n_min is not None:
        img_files = img_files[n_min:]   
    # create output folder
    if not os.path.exists(outfolder):
        os.makedirs(outfolder)
    # read in image - (mask certain area in image if specified)    
    if cut_img:
        img0 = io.imread(img_files[0], as_gray='True')[cut_img_val[0]:cut_img_val[1],  cut_img_val[2]:cut_img_val[3] ] 
    else:
        img0 = io.imread(img_files[0], as_gray='True')
              
    # segment , draw or load in mask
    if draw_mask == False:  # normal segmentation
        seg0 = segment_spheroid(img0, enhance=enhance, thres=thres_segmentation, thres_yen= thres_yen)
    else:                   # draw manual
        seg0 = custom_mask(img0)
    if load_mask is not None:   # load in mask
        seg0 = np.load(load_mask, allow_pickle=True).item()
   
    # save initial segmentation
    np.save(outfolder+'/seg000000.npy', seg0)
   
    # save parameter file for PIV calculation
    import yaml
    dict_file = {'PIV settings': {'folder': [folder], 'filter': [filter], 'outfolder': [outfolder], 
                  'n_max': [(n_max)], 'n_min': [(n_min)], 'enhance': [(enhance)], 
                  'window_size': [window_size], 'cutoff': [(cutoff)], 'drift_correction': [(drift_correction)], 
                  'drift_correction_raw_img': [(drift_correction_raw_img)], 
                  'plot': [(plot)], 'continous_segmentation': [(continous_segmentation)], 'quiver_scale': [quiver_scale], 
                  'color_norm': [color_norm], 'draw_mask': [(draw_mask)], 'gamma': [(gamma)], 
                  'gauss': [(gauss)], 'load_mask': [(load_mask)], 'thres_segmentation': [thres_segmentation], 
                  'cut_img': [(cut_img)], 'cut_img_val': [(cut_img_val)], 'cbar_um_scale': [cbar_um_scale], 
                  'dpi': [dpi], 'dt_min': [(dt_min)], 'cmap': [cmap], 'callback': [(callback)]},
                  'image_list': [image_list], 'Raw_img_files': [img_files], }
    with open(os.path.join(outfolder, 'parameters_piv.yml'), 'w') as yaml_file:
        yaml.dump(dict_file, yaml_file, default_flow_style=False)
    
    

    u_sum = None
    v_sum = None
    
    # loop through all images
    for i in tqdm(range(1, len(img_files))):
        # read in new image (and cut if specified)
        if cut_img:
            img1= io.imread(img_files[i], as_gray='True')[cut_img_val[0]:cut_img_val[1],  cut_img_val[2]:cut_img_val[3] ] 
        else:
            img1 = io.imread(img_files[i], as_gray='True')
        
        # take the mask of the current timestep if continous_segmentation is active -
        # if not always take the mask from t0 
        # both options have advantages and disadvantages: continous_segmentation might
        # determine the center over time more accurately and offers information over growth,
        # however wrongly detected masks here may lead to force fluctuations 
        # and can reduce the FoV too much. Since the standard approach for the 
        # later force reconstruction is unsing the initial timestep anyway, 
        # the approach of always using the t0 mask here as well is quite robust 
        if (draw_mask == False) & (continous_segmentation == True):
            seg1 = segment_spheroid(img1, enhance=enhance, thres=thres_segmentation)
        else:
            seg1 = seg0.copy()
        
        # compute and save the matrx deformations and mask    
        dis = compute_displacements(window_size, img0, img1, mask1=seg1['mask'],
                                cutoff=cutoff, drift_correction=drift_correction, drift_correction_raw_img=drift_correction_raw_img)
        np.save(outfolder + '/seg'+str(i).zfill(6)+'.npy', seg1)
        
        ## old conversion saves current deformation
        #np.save(outfolder + '/dis'+str(i).zfill(6)+'.npy', dis)  
                     
        
        if u_sum is None:
            u_sum = dis['u']
            v_sum = dis['v']
        else:
            u_sum += dis['u']
            v_sum += dis['v']

        dis_sum = {'x': dis['x'], 'y': dis['y'], 'u': u_sum, 'v': v_sum}
        ## new conversion saves cummulative deformations
        np.save(outfolder + '/def'+str(i).zfill(6)+'.npy', dis_sum)  
               
        
        # create timestamp if not None
        if dt_min is not None:
            t=dt_min*i
        else:
            t= None
                     
        if plot:   
            # plot same mask for all if continous segmentation is False or a custom mask is drawn
            if (draw_mask == False) & (continous_segmentation == True):
                save_displacement_plot(outfolder+'/plot'+str(i).zfill(6)+'.png', img1, seg1, dis_sum,
                                       quiver_scale=quiver_scale, color_norm=color_norm, colorlegend = colorlegend, cbar_um_scale=cbar_um_scale,dpi=dpi,cmap=cmap,t=t)
            # plot individual mask for each timestep ( however for forcereconstruction later we do use
            # the first timestep only to avoid errornous force fluctuations)    
            else:
                save_displacement_plot(outfolder+'/plot'+str(i).zfill(6)+'.png', img1, seg0, dis_sum,
                                       quiver_scale=quiver_scale, color_norm=color_norm, colorlegend = colorlegend, cbar_um_scale=cbar_um_scale,dpi=dpi,cmap=cmap,t=t)
        # next round we update the images        
        img0 = img1.copy()
        # call callback 
        if callback is not None:
            callback(i, len(img_files))



def compute_noise_level(folder, filter, n_max=10, enhance=True,
                        window_size=70, cutoff=None, drift_correction=True):

    img_files = natsorted(glob(folder + '/' + filter))

    if n_max is not None:
        img_files = img_files[:n_max+1]

    img0 = plt.imread(img_files[0])
    mask = np.zeros_like(img0).astype(np.bool)

    displ = []

    for i in tqdm(range(1, len(img_files))):
        img1 = plt.imread(img_files[i])
        seg1 = segment_spheroid(img1, enhance=enhance)

        dis = compute_displacements(window_size, img0, img1, mask1=mask,
                                    cutoff=cutoff, drift_correction=drift_correction)

        displ.extend(list(np.ravel(np.sqrt(dis['u']**2. + dis['v']**2.))))

        img0 = img1.copy()

    return np.percentile(displ, 95)



#!/usr/bin/env python

from setuptools import setup

setup(
    name='jointforces',
    packages=['jointforces'],
    version='1.0.3',
    description='A Python package to conduct 3D Traction Force Microcopy on multicelluar aggregates (spheroids).',
    url='',
    download_url = '',
    author='Christoph Mark, David Böhringer',
    author_email='christoph.mark@fau.de',
    license='The MIT License (MIT)',
    install_requires=['saenopy>=0.7.4',
                      'gmsh-sdk>=4.5.0',
                      'numpy>=1.16.2',
                      'pandas>=1.1.5',
                      'matplotlib>=2.2.2',
                      'scipy>=1.7.0',
                      'scikit-image>=0.14.2',
                      'openpiv>=0.23.4',
                      'tqdm>=4.26.0',
                      'natsort>=5.1.0', 
                      'dill>=0.2.9',
                      'roipoly>=0.5.2',
                      'openpyxl',
                      'pyyml',
					  'matplotlib-scalebar'],
    keywords = ['piv', 'contractility', 'material simulation', 'biophysics'],
    classifiers = [],
    )



def linear(stiffness):
    """
    Defines the stiffness for a linear elastic material.
    
    Args:
        stiffness(float) : Bulk modulus of the material in Pascal for SAENO Simulation (see [Steinwachs,2015])
    """
    return {'K_0': stiffness, 'D_0': 1e30, 'L_S': 1e30, 'D_S': 1e30}


def custom(K_0, D_0, L_S, D_S):
    """
    Defines the material properties for a custom nonlinear material.
    
    Args:
        K_0(float) : Bulk modulus of the material in Pascal for SAENO Simulation (see [Steinwachs,2015])
        D_0(float) : Buckling coefficient of the fibers for SAENO Simulation (see [Steinwachs,2015])
        L_S(float) : Onset of strain stiffening of the fibers (see [Steinwachs,2015])
        D_S(float) : Strain stiffening coefficient of the fibers (see [Steinwachs,2015])   
    """
    return {'K_0': K_0, 'D_0': D_0, 'L_S': L_S, 'D_S': D_S}


# The following material parameters are taken from Steinwachs et al. (2016).
# See https://www.nature.com/articles/nmeth.3685 for detailed protocols to
# replicate the bioploymer gels - 

# 1:1 mixture of Collagen G and Collagen R (0.6mg/ml)
collagen06 = {'K_0': 447, 'D_0': 0.0008, 'L_S': 0.0075, 'D_S': 0.033}

# 1:1 mixture of Collagen G and Collagen R (1.2mg/ml)
collagen12 = {'K_0': 1645, 'D_0': 0.0008, 'L_S': 0.0075, 'D_S': 0.033}

# 1:1 mixture of Collagen G and Collagen R (2.4mg/ml)
collagen24 = {'K_0': 5208, 'D_0': 0.0008, 'L_S': 0.0075, 'D_S': 0.033}

# Fibrin gel (4.0mg/ml)
fibrin40 = {'K_0': 2091, 'D_0': 0.002, 'L_S': 1e30, 'D_S': 1e30}

# Matrigel (10mg/ml)
matrigel100 = {'K_0': 2364, 'D_0': 1e30, 'L_S': 1e30, 'D_S': 1e30}
